
from __future__ import unicode_literals
import os, sys, re
import codecs
import inspect
import uuid
import datetime

KEY_ORDER_DEFAULT = 10000
LINKED_ART_CONTEXT_URI = "https://linked.art/ns/v1/linked-art.json"

# 2.5 and 2.6 are very out of date. Assume 2.7 or better
import json
from collections import OrderedDict, namedtuple

from json import JSONEncoder

try:
	STR_TYPES = [str, unicode] #Py2.7
	FILE_STREAM_CLASS = file
except:
	import io
	STR_TYPES = [bytes, str] #Py3.x
	FILE_STREAM_CLASS = io.TextIOBase


PropInfo = namedtuple("PropInfo", [
	'property', # the name of the property, eg 'identified_by'
	'predicate', # the name from the ontology, eg 'crm:P1_is_identified_by'
	'range', # the class which is the range of this predicate, eg model.Identifier
	'inverse_property', # the name of the property which is the inverse of this one, eg 'identifies'
	'inverse_predicate', # the predicate for the inverse of this one, eg 'crm:P1i_idenfifies'
	'multiple_okay', # can there be multiple values, eg True
	'profile_okay' # is this property okay according to the profile, eg 0 (no), 1 (yes), 2 (warn)
	])


class CromulentError(Exception):
	"""Base exception class"""

	resource = None

	def __init__(self, msg, resource=None):
		"""Initialize CidocError."""
		self.args = [msg]
		self.resource = resource

class ConfigurationError(CromulentError):
	"""Raised when an object (likely the factory) isn't configured properly for the current operation."""
	pass

class MetadataError(CromulentError):
	"""Base metadata exception."""
	pass

class RequirementError(MetadataError):
	"""Raised when schema/profile/metadata requirements not met."""
	pass

class DataError(MetadataError):
	"""Raised when data is not valid/allowed."""
	pass

class ProfileError(MetadataError):
	"""Raised when a class or property not in the configured profile is used"""
	pass


class CromJsonEncoder(JSONEncoder):

	def default(self, o):
		if isinstance(o, BaseResource):
			# print("Saw %r" % o)
			return o._minToJSON()
		else:
			return JSONEncoder.default(self, o)			


class CromulentFactory(object):

	def __init__(self, base_url="", base_dir="", lang="", full_names=False, 
		context="", context_file={}, load_context=True):
		self.base_url = base_url
		self.base_dir = base_dir

		self.debug_level = "warn"
		self.log_stream = sys.stderr

		self.materialize_inverses = False # Create the inverse relationship on the object to the subject
		self.full_names = False # Use full property and class names in output
		self.pipe_scoped_contexts = False # Serialize to "name|Pnn_x_y" for scoped properties (documentation)
		self.validate_properties = True # validate properties are even CRM
		self.validate_profile = True # Validate linked_art specific stuff
		self.validate_range = True # Raise if attempt to set prop to invalid class
		self.validate_multiplicity = True  # Raise if attempt to set n:1 to []
		self.auto_assign_id = True # Automatially assign a URI
		self.process_multiplicity = True # Return multiple with single value as [value]
		self.multiple_instances_per_property = "drop"
		self.allow_highlight = False # Allow the JSON to include a _highlight flag for re-rendering
		self.allow_elide = False

		self.auto_id_type = "int-per-segment" #  "int", "int-per-type", "int-per-segment", "uuid"
		# self.default_lang = lang  # NOT USED
		self.filename_extension = ".json"  # some people like .jsonld
		self.context_uri = context # Might be a list, or a context value as a dict
		self.context_json = {}

		self.prefixes = {}
		self.prefixes_rev = {}
		self.context_rev = {}
		# Maybe load it up for prefixes
		if load_context:
			# Leave this as a map for future extensions
			context_filemap = {
				LINKED_ART_CONTEXT_URI: 
					os.path.join(os.path.dirname(__file__), 'data', 'linked-art.json')
			}
			context_filemap.update(context_file)
			self.load_context(context, context_filemap)

		self.elasticsearch_compatible = False # return {'id': 'uri'} instead of string
		self.linked_art_boundaries = False # break on linked art API boundaries between classes
		self.id_type_label = True # references are id, type and _label, not just id.

		# if sorting is unimportant, use fast. If sorting is important, and python >= 3.6, use fast.
		# fast is approximately half the time for serializing
		self.json_serializer = "normal" # "normal" or "fast"
		self.json_indent = 2
		self.order_json = True
		self.key_order_hash = {"@context": 0, "id": 1, "type": 2, 
			"_label": 5, "value": 6}
		self.full_key_order_hash = {"@context": 0, "@id": 1, "rdf:type": 2, "@type": 2,
			"rdfs:label": 5, "rdf:value": 6}
		self.key_order_default = 10000

		self.underscore_properties = ["_label"]

		self._auto_id_types = {}
		self._auto_id_segments = {}
		self._auto_id_int = -1
		self._all_classes = {}

	def load_context(self, context, context_filemap):
		if not context or not context_filemap:
			raise ConfigurationError("No context provided, and load_context not False")

		if type(context) is not list:
			context = [context]

		js = {'@context': {}}
		for ct in context:
			fn = context_filemap.get(ct, "")
			if fn:
				try:
					fh = open(fn)
					data = fh.read()
					fh.close()
				except IOError:
					raise ConfigurationError("Provided context file does not exist")
			else:
				# XXX Fetch from web
				data = "{}"

			try:
				ctx = json.loads(data)
				js['@context'].update(ctx['@context'])
			except:
				raise ConfigurationError("Provided context does not have valid JSON")				
		# this is the merged context information, not any single one
		self.context_json = js
		self.process_context()

	def process_context(self):
		# Filter context looking for prefixes
		# And make reverse mapping
		# Note that this does not process scoped contexts for member_of / part_of
		for (k,v) in self.context_json['@context'].items():
			if type(v) in STR_TYPES and v[-1] in ['/', '#']:
				self.prefixes[k] = v
				self.prefixes_rev[v] = k
			elif k == "@version":
				continue
			else:
				if type(v) in STR_TYPES:
					rdf = v
				else:
					rdf = v['@id']
				self.context_rev[rdf] = k

	def __getstate__(self):
		# Make a copy of current object state
		d = self.__dict__.copy()
		# try to flush the stream
		try:
			self.log_stream.flush()
		except:
			# stream instance may not support flush()
			pass
		# Now reify the log stream
		strm = d['log_stream']
		if strm is sys.stdout:
			d['log_stream'] = ("sys.stdout", "stream")
		elif strm is sys.stderr:
			d['log_stream'] = ("sys.stderr", "stream")
		elif isinstance(strm, FILE_STREAM_CLASS):
			d['log_stream'] = (strm.name, "file")
		else:
			d['log_stream'] = None
		return d

	def __setstate__(self, state):
		# State is __dict__ with a reified log_stream as above
		self.__dict__.update(state)
		if self.log_stream:
			if self.log_stream[1] == "stream":
				if self.log_stream[0] == "sys.stdout":
					self.log_stream = sys.stdout
				elif self.log_stream[0] == "sys.stderr":
					self.log_stream = sys.stderr
			elif self.log_stream[1] == "file":
				try:
					self.log_stream = open(self.log_stream[0], 'a') 
				except:
					self.log_stream = None


	def set_debug_stream(self, strm):
		"""Set debug level."""
		self.log_stream = strm

	def set_debug(self, typ):
		"""Set behavior on errors and warnings.

		error = squash warnings
		warn = display warnings
		error_on_warning = raise exception for a warning rather than continuing
		"""
		if typ in ['error', 'warn', 'error_on_warning']:
			self.debug_level = typ
		else:
			raise ConfigurationError("Only levels are 'error', 'warn' and 'error_on_warning'")

	def maybe_warn(self, msg):
		"""warn method that respects debug_level property."""
		if self.log_stream and self.debug_level == "warn":
			self.log_stream.write(msg + "\n")
			try:	
				self.log_stream.flush()
			except:
				pass
		elif self.debug_level == "error_on_warning":
			# We don't know the type, just raise a MetadataError
			raise MetadataError(msg)

	def _is_uri(self, what):
		uri_schemes = ['urn:uuid:', 'tag:', 'data:', 'mailto:', 'info:', 'ftp:/', 'sftp:/'] 
		for u in uri_schemes:
			if what.startswith(u):
				return True
		return False

	def generate_id(self, what, auto_type=None):
		if auto_type == None:
			auto_type = self.auto_id_type

		if auto_type == "int":
			# increment and return
			self._auto_id_int += 1
			slug = self._auto_id_int
		elif auto_type == "int-per-segment":
			curr = self._auto_id_segments.get(what._uri_segment, -1)
			curr += 1
			self._auto_id_segments[what._uri_segment] = curr
			slug = self._auto_id_segments[what._uri_segment]
		elif auto_type == "int-per-type":
			t = type(what).__name__
			curr = self._auto_id_types.get(t, -1)
			curr += 1
			self._auto_id_types[t] = curr
			slug = self._auto_id_types[t]
		elif auto_type == "uuid":
			return "urn:uuid:%s" % uuid.uuid4()
		else:
			raise ConfigurationError("Unknown auto-id type")

		return self.base_url + what.__class__._uri_segment + "/" + str(slug)		

	def toJSON(self, what, done=None):
		""" Serialize what, making sure of no infinite loops """
		if not done:
			done = {}
		if self.json_serializer == "fast":
			out = what._toJSON_fast(top=what, done=done)
		else:
			out = what._toJSON(top=what, done=done)
		return out

	def _collapse_json(self, text, collapse):
		js_indent = self.json_indent
		lines = text.splitlines()
		out = [lines[0]]
		while lines:
			l = lines.pop(0)
			indent = len(re.split('\S', l, 1)[0])
			if indent and l.rstrip()[-1] in ['[', '{']:
				curr = indent
				temp = []
				stemp = []
				while lines and curr <= indent:
					if temp and curr == indent:
						break
					temp.append(l[curr:])
					stemp.append(l.strip())
					l = lines.pop(0)
					indent = len(re.split('\S', l, 1)[0])					
				temp.append(l[curr:])
				stemp.append(l.lstrip())

				short = " " * curr + ''.join(stemp)
				if len(short) < collapse:
					out.append(short)
				else:
					ntext = '\n'.join(temp)
					nout = self._collapse_json(ntext, collapse)					
					for no in nout:
						out.append(" " * curr + no)
			elif indent:
				out.append(l)
		out.append(l)
		return out

	def collapse_json(self, text, collapse):
		return '\n'.join(self._collapse_json(text, collapse))

	def _buildString(self, js, compact=True, collapse=0):
		"""Build string from JSON."""
		try:
			if compact:
				out = json.dumps(js, separators=(',',':'), ensure_ascii=False)
			else:
				out = json.dumps(js, indent=self.json_indent, ensure_ascii=False)
		except:
			out = ""
			self.maybe_warn("Can't decode %r" % js)
			raise
		if collapse:
			out = self.collapse_json(out, collapse)
		return out 		

	def toString(self, what, compact=True, collapse=0, done=None):
		"""Return JSON setialization as string."""
		if not done:
			done = {}
		js = self.toJSON(what, done=done)
		return self._buildString(js, compact, collapse)


	def toHtml(self, what, done=None):
		enc = JSONEncoder(indent=self.json_indent, ensure_ascii=False)
		js = self.toJSON(what, done=done)
		res = ['<pre><span>']
		lineno = 0
		for y in enc.iterencode(js):
			# split newlines
			if '\n' in y:
				sp = y.split('\n')
				x2 = []
				for z in sp[:-1]:
					if z:
						x2.append(z)
					x2.append('\n')
				if sp[-1]:
					x2.append(sp[-1])
			else:
				x2 = [y]

			for x in x2:
				if x[0] == '"':
					s = x.strip()[1:-1]
					clss = "str"
					if s.startswith('http') or self._is_uri(s):
						x = '"<a href="%s">%s</a>"' % (s, s)
					elif s in ["@context"]:
						clss = "str context"
					res.append('<span class="%s">%s</span>' % (clss, x))
				elif x[0].isdigit():
					res.append('<span class="int">%s</span>' % x)						
				elif x[0] in [":", ","]:
					res.append('<span class="sep">%s</span>' % x)	
				elif x[0]  == "{":
					res.append('<span class="bsep">%s</span><span class="block">' % x)
				elif x[0]  == "[":
					res.append('<span class="asep">%s</span><span class="ablock">' % x)
				elif x[0] == "}":
					res.append('</span><span class="bsep">%s</span>' % x)
				elif x[0] == "]":
					res.append('</span><span class="asep">%s</span>' % x)
				elif x == '\n':
					res.append('\n<span class="line" data-lno="%s"></span>' % lineno)
					lineno += 1
				elif x.isspace():
					res.append(x)
				else:
					print("*** Unhandled: %r" % x)
					res.append(x)
		res.append('</span></pre>')
		return ''.join(res)


	def toFile(self, what, compact=True, filename="", done=None):
		"""Write to local file.

		Creates directories as necessary
		"""

		if not done:
			done = {}
		js = self.toJSON(what, done=done)

		if not filename:
			myid = js['id']
			mdb = self.base_url
			if not myid.startswith(mdb):
				raise ConfigurationError("The id of that object is not the base URI in the Factory")
			mdd = self.base_dir
			if not mdd:
				raise ConfigurationError("Directory on Factory must be set to write to file")
			fp = myid[len(mdb):]	
			bits = fp.split('/')
			if len(bits) > 1:
				mydir = os.path.join(mdd, '/'.join(bits[:-1]))		
				try:
					os.makedirs(mydir)
				except OSError:
					pass
			if self.filename_extension:
				fp = fp + self.filename_extension
			filename = os.path.join(mdd, fp)

		fh = open(filename, 'w')
		out = self._buildString(js, compact)
		fh.write(out)
		fh.close()
		return out

	def production_mode(self, state=True):
		if state:
			self.cache_hierarchy()
			self.validate_profile = False
			self.validate_properties = False
			self.validate_range = False
			self.validate_multiplicity = False
			return True
		else:
			self.validate_profile = True
			self.validate_properties = True
			self.validate_range = True
			self.validate_multiplicity = True
			# no way to undo the hierarchy cache
			return False
			
	def cache_hierarchy(self):
		""" For each class, walk up the hierarchy and cache the terms """
		# This will work with the existing code, as it will find it in the first
		# test of props on classhier[0], the loop will just terminate straight away

		for c in self._all_classes.values():
			new_hash = c._all_properties.copy()
			if len(c._classhier) > 1:
				for p in c._classhier[1:]:
					for (prop, info) in p._all_properties.items():
						if not prop in new_hash:
							new_hash[prop] = info
			c._all_properties = new_hash


class ExternalResource(object):
	"""Base class for all resources, including external references"""
	
	_factory = None
	_uri_segment = ""
	id = ""
	_full_id = ""
	_all_properties = {}
	_type = ""
	_embed = True
	_property_name_map = {}
	_highlight = False
	_elide = False


	def __init__(self, ident=None):
		self._factory = factory
		if ident is not None:
			if self._factory._is_uri(ident):
				self.id = ident
			elif ident.startswith('http'):
				# Try to find prefixable term
				hashed = ident.rsplit('#', 1)
				if len(hashed) == 1:
					(pref, rest) = ident.rsplit('/', 1)
					pref += "/"
				else:
					(pref, rest) = hashed
					pref += "#"

				if pref in self._factory.prefixes_rev:
					self._full_id = ident
					ident = "%s:%s" % (self._factory.prefixes_rev[pref], rest)

				self.id = ident
			elif ident == "":
				# Allow explicit setting of empty string
				self.id = ""
			elif ident.startswith("auto "):
				# Allow override of auto setting for this resource
				# eg:  model.Person(ident="auto int") to assign an int, rather than the default
				self.id = factory.generate_id(self, auto_type=ident[5:].strip())
			else:
				# Allow for prefixed term that isn't ambiguously a URI
				curied = ident.split(':', 1)
				if len(curied) == 2 and curied[0] in self._factory.prefixes:
					self.id = ident
					self._full_id = self._factory.prefixes[curied[0]] + curied[1]	
				else:
					self.id = factory.base_url + self.__class__._uri_segment + "/" + ident
		elif factory.auto_assign_id:
			self.id = factory.generate_id(self)
		else:
			# Not auto assigning, and not submitted = blank node
			self.id = ""

	def _toJSON(self, done, top=None):
		if self._factory.elasticsearch_compatible:
			return {'id': self.id}
		else:
			return self.id

class BaseResource(ExternalResource):
	"""Base class for all resources with classes"""

	_integer_properties = []
	_object_properties = []
	_required_properties = []
	_warn_properties = []
	_classification = ""
	_classhier = []

	def __init__(self, ident=None, label="", value="", content="", **kw):
		"""Initialize BaseObject."""
		super(BaseResource, self).__init__(ident)

		if self._factory.validate_profile and hasattr(self, '_okayToUse'): 
			if not self._okayToUse:
				raise ProfileError("Class '%s' is configured to not be used" % self.__class__._type)
			elif self._okayToUse == 2:
				self._factory.maybe_warn("Class '%s' is configured to warn on use" % self.__class__._type)

		# Set label and value/content
		if label:
			self._label = label
		# this might raise an exception if value is not allowed on the object
		# but easier to do it in the main init than on many generated subclasses

		is_sym = isinstance(self, SymbolicObject)
		is_dim = isinstance(self, Dimension)

		if value and is_dim:
			self.value = value
		elif content and is_sym:
			self.content = content
		elif value and is_sym:
			self.content = value # not the right param, but not ambiguous
		elif content and is_dim:
			self.value = content # ditto
		elif value or content: 
			raise ProfileError("Class '%s' does not hold values" % self.__class__._type)
		# Custom post initialization function for autoconstructed classes
		self._post_init(**kw)

	def __dir__(self):
		d = dir(self.__class__)
		d.extend(self.list_all_props())
		return sorted(d)

	def __eq__(a, b):
		if id(a) == id(b):
			return True
		if not a or not b:
			return False
		if not isinstance(b, BaseResource) or not isinstance(a, BaseResource):
			return False
		ap = a.list_my_props()
		bp = b.list_my_props()
		if ap != bp:
			return False
		for p in ap:
			av = getattr(a, p)
			bv = getattr(b, p)
			if av != bv:
				return False
		return True

	@property
	def type(self):
		for c in self._classhier:
			if c._type:
				return c.__name__

	@type.setter
	def type(self, value):
		raise AttributeError("Must not set 'type' on resources directly")
	
	def set_context(self, value):	
		raise DataError("Must not set the JSON LD context directly", self)

	def _post_init(self, **kw):
		# Expect this to be overridden / replaced
		pass

	def __setattr__(self, which, value):
		"""Attribute setting magic for error checking and resource/literal handling."""

		if which[0] == "_" or not value:
			# _label goes through here, but it would below anyway, as it takes a Literal
			object.__setattr__(self, which, value)			
		else:
			# Allow per-class setters
			if hasattr(self, 'set_%s' % which):
				fn = getattr(self, 'set_%s' % which)
				return fn(value)

			if self._factory.validate_properties or self._factory.validate_profile or self._factory.validate_range:
				ok = self._check_prop(which, value)
			elif isinstance(value, ExternalResource):
				ok = 2
			else:
				ok = 1

			if ok == 2:
				self._set_magic_resource(which, value)
			else:			
				object.__setattr__(self, which, value)				
		 
	def _check_prop(self, which, value):
		val_props = self._factory.validate_properties
		val_profile = self._factory.validate_profile
		val_range = self._factory.validate_range
		for c in self._classhier:
			if which in c._all_properties:
				pinfo = c._all_properties[which]

				if val_profile:
					okay = pinfo.profile_okay					
					rdf = pinfo.predicate
					if not okay:
						raise ProfileError("Property '%s' / '%s' is configured to not be used" % (which, rdf), self)
					elif okay == 2:
						self._factory.maybe_warn("Property '%s' / '%s' is configured to warn on use" % (which, rdf))

				if val_range:
					rng = pinfo.range
					if rng is str:					
						return 1
					elif type(value) is BaseResource:
						# Allow direct instances of base resource anywhere
						# this is an override for external URIs
						return 2
					elif isinstance(value, rng):
						return 2
					else:
						raise DataError("Can't set '%s' on resource of type '%s' to '%r'" % (which, self._type, value), self)
				# Found it, but not validating range and either okay or not validating profile
				return 1
		if val_props:
			raise DataError("Can't set unknown field '%s' on resource of type '%s'" % (which, self._type), self)
		else:
			# Not validating ANYTHING
			return 1

	def _check_reference(self, data):
		"""True if data is a resource or reference to a resource"""
		# "http://..."
		# {"@id": "http://..."}
		# or list of above
		if type(data) in STR_TYPES:
			return data.startswith('http')
		elif type(data) is dict:
			return 'id' in data
		elif isinstance(data, BaseResource):
			return True
		elif type(data) is list:
			for d in data:
				if type(d) in STR_TYPES and not d.startswith('http'):
					return False
				elif type(d) is dict and not 'id' in d:
					return False
			return True
		else:
			self._factory.maybe_warn("expecing a resource, got: %r" % (data))
			return True
				
	def _set_magic_resource(self, which, value, inversed=False):
		"""Set resource property.
		allow: string/object/dict, and magically generate list thereof
		"""

		if self._factory.materialize_inverses or self._factory.process_multiplicity or \
			self._factory.validate_multiplicity:
			inverse = None
			multiple = 1
			for c in self._classhier:
				if which in c._all_properties:
					v = c._all_properties[which]
					multiple = v.multiple_okay
					if v.inverse_property:
						inverse = v.inverse_property
					break

		try:
			current = getattr(self, which)
		except:
			current = None
		if not current:
			object.__setattr__(self, which, value)
		elif type(current) is list:
			# check value not in list
			if self._factory.multiple_instances_per_property == "error" and isinstance(value, BaseResource) and value in current:
				raise DataError("""Cannot add the same resource in the same property more than once:
change factory.multiple_instances_per_property to 'drop' or 'allow'""")
			current.append(value)
		else:
			if self._factory.validate_multiplicity and not multiple:
				raise ProfileError("Cannot append to %s on %s as multiplicity is 1" % (which, self._type))
			nvalue = [current, value]
			object.__setattr__(self, which, nvalue)

		if self._factory.materialize_inverses and not inversed and inverse:
			# set the backwards ref		
			value._set_magic_resource(inverse, self, True)
		if self._factory.process_multiplicity and type(current) is not list and multiple:
			object.__setattr__(self, which, [getattr(self, which)])


	def _toJSON(self, done, top=None):
		"""Serialize as JSON."""
		# If we're already in the graph, return our URI only
		# This should only be called from the factory!

		d = self.__dict__.copy()
		del d['_factory']

		# Can't pass in self as a param
		if top is None:
			top = self

		# id, type, _label is the default.
		if not factory.id_type_label and id(self) in done:
			if self._factory.elasticsearch_compatible:
				return {'id': self.id}
			else:
				return self.id

		# In case of local contexts, not at the root
		# Shouldn't ever happen, but worth testing for
		if 'context' in d:
			d['@context'] = d['context']
			del d['context']

		# Check mandatory properties
		if self._factory.validate_profile:
			for e in self._required_properties:
				if e not in d:
					raise RequirementError("Resource type '%s' requires '%s' to be set" % (self._type, e), self)

		debug = self._factory.debug_level
		if debug.find("warn") > -1:
			for e in self._warn_properties:
				if e not in d:
					msg = "Resource type '%s' should have '%s' set" % (self._type, e)
					self._factory.maybe_warn(msg)

		# Add back context at the top, if set
		if top is self and self._factory.context_uri: 
			d['@context'] = self._factory.context_uri

		if (self._factory.id_type_label and id(self) in done) or (top is not self and not self._embed):
			# limit to only id, type, label
			nd = {}
			nd['id'] = d['id']
			if self.type:
				nd['type'] = self.type
			try:
				nd['_label'] = d['_label']
			except:
				pass
			d = nd
		else:	
			# otherwise, we're about to serialize the resource completely
			done[id(self)] = 1			

		# Need to do in order now to get done correctly ordered
		if self._factory.order_json:
			KOH = self._factory.key_order_hash
			kodflt = self._factory.key_order_default
			kvs = sorted(d.items(), key=lambda x: KOH.get(x[0], kodflt))
		else:
			kvs = list(d.items())

		tbd = []
		for (k, v) in kvs:
			# some _foo might be carried through, eg _label or _comment
			k = self._property_name_map.get(k, k)
			if not v or (k[0] == "_" and not k in self._factory.underscore_properties):
				del d[k]
			else:

				# Should we do this at all? Could be outside of our API serialization scope
				if isinstance(v, ExternalResource):
					if self._factory.linked_art_boundaries and \
						not self._linked_art_boundary_okay(top, k, v):
						# never follow, so just add to done
						done[id(v)] = 1
					else:
						tbd.append(id(v))
				elif type(v) is list:
					for ni in v:
						if isinstance(ni, ExternalResource):
							if self._factory.linked_art_boundaries and \
								not self._linked_art_boundary_okay(top, k, ni):
								# never follow, so just add to done
								done[id(ni)] = 1							
							else:
								tbd.append(id(ni))
					# For completeness should check list-of-datetime here too
				elif isinstance(v, datetime.datetime):
					# replace with string
					kvs[k] = v.strftime("%Y-%m-%dT%H:%M:%SZ")

		for t in tbd:
			if not t in done:
				done[t] = id(self)
			
		for (k,v) in kvs:
			nk = self._property_name_map.get(k, k)
			if nk != k:
				del d[k]
				d[nk] = v
				k = nk
			if v and (k[0] != "_" and not k in self._factory.underscore_properties):
				if isinstance(v, ExternalResource):
					if done[id(v)] == id(self):
						del done[id(v)]
					d[k] = v._toJSON(done=done, top=top)
				elif type(v) is list:
					newl = []
					uniq = set()
					for ni in v:
						if self._factory.multiple_instances_per_property == "drop":
							if id(ni) in uniq:
								continue
							else:
								uniq.add(id(ni))
						if isinstance(ni, ExternalResource):
							if done[id(ni)] == id(self):
								del done[id(ni)]
							newl.append(ni._toJSON(done=done, top=top))
						else:
							# A number or string
							newl.append(ni)
					d[k] = newl				

		if self._factory.full_names:
			nd = {}
			# @context gets ganked by this renaming
			# so add it back in first.
			if top is self:
				nd['@context'] = self._factory.context_uri

			for (k,v) in d.items():
				# look up the rdf predicate in _properties
				for c in reversed(self._classhier):
					if k in c._all_properties:
						nk = c._all_properties[k].predicate
						nd[nk] = v
						break

			# Ensure full version uses basic @type
			if "rdf:type" in nd:
				nd['@type'] = nd['rdf:type']
				del nd['rdf:type']

			# And type gets ganked for overlay classes (Painting)
			# plus for stupidity classes (DestructionActivity)
			# so add this back too
			if not "@type" in nd or not nd['@type']:
				# find class up that has a type and use its name
				for c in reversed(self._classhier):
					if c._type:
						nd['@type'] = c._type

			d = nd
			KOH = self._factory.full_key_order_hash
		else:
			# Use existing programmer-friendly names for classes too
			# find class up that has a type and use its name
			# almost certainly the first one
			if self.type:
				d['type'] = self.type

			if self._factory.pipe_scoped_contexts:
				# XXX TODO This should be configurable not hard coded
				if 'part' in d:
					# Calculate which part
					for c in reversed(self._classhier):
						if 'part' in c._all_properties:
							nk = c._all_properties['part'].predicate
							d['part|%s' % nk]  = d['part']
							del d['part']
							break
				if 'part_of' in d:
					# Calculate which part
					for c in reversed(self._classhier):
						if 'part_of' in c._all_properties:
							nk = c._all_properties['part_of'].predicate
							d['part_of|%s' % nk]  = d['part_of']
							del d['part_of']
							break

		if self._highlight and self._factory.allow_highlight:
			d['_highlight'] = True
		if self._elide and self._factory.allow_elide:
			d['_elide'] = True

		if self._factory.order_json:
			return OrderedDict(sorted(d.items(), key=lambda x: KOH.get(x[0], 1000)))
		else:
			return d

	def _toJSON_fast(self, done, top=None):
		"""Serialize as JSON."""
		# If we're already in the graph, return our URI only
		# This should only be called from the factory!

		# This should ONLY be used in Python3.6+
		# as it relies on a stable order to a regular dict
		# which was not the case before 3.6

		# id, type, _label is the default.
		if not self._factory.id_type_label and id(self) in done:
			return self.id

		# Can't pass in self as a param
		if top is None:
			top = self

		# Add back context at the top, if set
		result = {}
		if top is self and id(self) not in done and self._factory.context_uri: 
			result['@context'] = self._factory.context_uri

		if self.id:
			# Could be a bnode
			result['id'] = self.id
		if self.type:
			# Could be an external reference
			result['type'] = self.type
		try:
			result['_label'] = self._label
		except:
			pass

		# Need only minimal representation of self
		if (self._factory.id_type_label and id(self) in done) or (top is not self and not self._embed):
			# limit to only id, type, label
			return result
		else:	
			# otherwise, we're about to serialize the resource completely
			done[id(self)] = 1			

		d = self.__dict__.copy()
		del d['_factory']
		del d['id']

		# Need to do in order now to get done correctly ordered
		kvs = list(d.items())
		if self._factory.order_json:
			KOH = self._factory.key_order_hash
			kodflt = self._factory.key_order_default
			# in place sort is 2% faster than sorted
			# rather than over an iterable
			# and sorting is 10% of srlz cost
			# and use setdefault to amortize cost of default
			# over multiple calls to the factory's hash
			kvs.sort(key=lambda x: KOH.setdefault(x[0], kodflt))
			
		# tbd vs done is to ensure that in a DAG rather than a tree
		# that children of the current node are given priority for
		# full serialization, rather than deeper descendent nodes
		# that would otherwise be processed first during the (recursive)
		# depth-first traversal.

		# This doesn't catch the pattern A-B-C-D / A-E-D,
		# (D will be under B-C, not under E) as B is processed completely
		# before E.

		# Note: Reimplemented this in a single depth-first loop
		# and it was less than 2% faster. The 2 loops are not that expensive

		tbd = []
		for (k, v) in kvs:
			k = self._property_name_map.get(k, k)
			if isinstance(v, ExternalResource):
				if self._factory.linked_art_boundaries and \
					not self._linked_art_boundary_okay(top, k, v):
					# never follow, so just add to done
					done[id(v)] = 1
				else:
					tbd.append(id(v))
			elif type(v) is list:
				for ni in v:
					if isinstance(ni, ExternalResource):
						if self._factory.linked_art_boundaries and \
							not self._linked_art_boundary_okay(top, k, ni):
							# never follow, so just add to done
							done[id(ni)] = 1							
						else:
							tbd.append(id(ni))

		for t in tbd:
			if not t in done:
				done[t] = id(self)
			
		# This is already sorted if needed
		for (k,v) in kvs:
			k = self._property_name_map.get(k, k)
			if not v:
				pass
			elif isinstance(v, ExternalResource):
				if done[id(v)] == id(self):
					del done[id(v)]
				result[k] = v._toJSON_fast(done=done, top=top)
			elif type(v) is list:
				newl = []
				uniq = set()
				for ni in v:
					if self._factory.multiple_instances_per_property == "drop":
						if id(ni) in uniq:
							continue
						else:
							uniq.add(id(ni))
					if isinstance(ni, ExternalResource):
						if done[id(ni)] == id(self):
							del done[id(ni)]
						newl.append(ni._toJSON_fast(done=done, top=top))
					else:
						# A number or string
						newl.append(ni)
				result[k] = newl
			elif isinstance(v, datetime.datetime):
				result[k] = v.strftime("%Y-%m-%dT%H:%M:%SZ")
			else:
				result[k] = v
		return result

	def _linked_art_boundary_okay(self, top, prop, value):
		# Return false to say do not cross this boundary
		# Without replacement, just return True always
		return True

	def list_all_props(self, filter=None, okay=None):
		props = []
		for c in self._classhier:		
			for k,v in c._all_properties.items():
				if not k in props and \
					(not okay or (okay and v.profile_okay)) and \
					(filter is None or isinstance(filter, v.range) or \
						filter is v.range):
					props.append(k)
		props.sort()
		return props

	def list_all_props_with_range(self, filter=None, okay=None):
		props = {}
		for c in self._classhier:		
			for k,v in c._all_properties.items():
				if not k in props and \
					(not okay or (okay and v.profile_okay)) and \
					(filter is None or isinstance(filter, v.range) or \
						filter is v.range):
					props[k] = v.range
		return props

	def list_my_props(self, filter=None):
		d = self.__dict__.copy()		
		props = []
		for (k,v) in d.items():
			if k[0] != "_" or k in self._factory.underscore_properties:
				# real property
				if filter:
					if isinstance(v, filter):
						props.append(k)
					elif isinstance(v, list):
						for i in v:
							if isinstance(i, filter):
								props.append(k)
								break
				else:
					props.append(k)		
		return props

	def allows_multiple(self, propName):
		""" Does propName allow multiple values on this class """
		for c in self._classhier:
			if propName in c._all_properties:
				v = c._all_properties[propName]
				return bool(v.multiple_okay)
		raise DataError("Cannot set '%s' on '%s'" % (propName, self.__class__.__name__))

	def clone(self, minimal=False):
		""" Make a shallow copy of self. If minimal flag is set, only return id, type and _label """
		new = self.__class__(ident=self.id)
		if minimal:
			if self._label:
				new._label = self._label
		else:
			for p in self.list_my_props():
				curr = getattr(self, p)
				if type(curr) == list:
					for v in curr:
						setattr(new, p, v)
				else:
					setattr(new, p, curr)
		return new

# Ontology / Profile manipulation

def override_okay(clss, propName):
	""" set particular property on the class to be okay to use """
	pinfo = clss._all_properties.get(propName, None)
	if pinfo:
		npinfo = PropInfo(pinfo.property, pinfo.predicate, pinfo.range,
			pinfo.inverse_property, pinfo.inverse_predicate, 
			pinfo.multiple_okay, 1)
		clss._all_properties[propName] = npinfo
	else:
		raise DataError("%s does not have a %s property to allow" % 
			(clss.__name__, propName))


# Ensure everything can have id, type, label and description
BaseResource._properties = {'id': {"rdf": "@id", "range": str, "okayToUse": 1}, 
	'type': {"rdf": "rdf:type", "range": str, "rangeStr": "rdfs:Class", "okayToUse": 1}, 
	'_label': {"rdf": "rdfs:label", "range": str, "rangeStr": "xsd:string", "okayToUse": 1}
}
BaseResource._all_properties = {
	'id': PropInfo('id', '@id', str, None, None, 0, 1),
	'type': PropInfo('type', 'rdf:type', str, None, None, 0, 1),
	'_label': PropInfo('_label', 'rdfs:label', str, None, None, 0, 1)
}
BaseResource._classhier = (BaseResource, ExternalResource)

def process_tsv(fn):
	fh = codecs.open(fn, 'r', 'utf-8')
	lines = fh.readlines()[1:] # chomp header line
	fh.close()
	vocabData = {"rdf:Resource": 
		{"props": [], "label": "Resource", "className": "Resource", 
		"subs":[], "desc": "", "class": BaseResource, "okay": 1}}

	for l in lines:
		l = l[:-1] # chomp
		info= l.split('\t')
		name = info[0]	
		if info[1] == "class":
			data = {"subOf": info[5], "label": info[3], 'className': info[2],
				"desc": info[4], "class": None, "props": [], "subs": [], "okay": info[6]}
			vocabData[name] = data
		else:
			# property
			data = {"name": name, "subOf": info[5], "label": info[3], "propName": info[2],
			"desc": info[4], "range": info[7], "inverse": info[8], "okay": info[10], "multiple": info[11]}
			try:
				what = vocabData[info[6]]
			except:
				print("Failed to find class for %s given %s" % (data, info[6]))
				raise
			what["props"].append(data)

			koh = int(info[9])
			if koh != KEY_ORDER_DEFAULT:
				factory.key_order_hash[data['propName']] = koh
				factory.full_key_order_hash[data['name']] = koh

	# invert subclass hierarchy
	for k, v in vocabData.items():
		if k != "rdf:Resource":		
			sub = v['subOf']
			# | separated list
			for s in sub.split('|'):
				if s:
					try:
						vocabData[s]['subs'].append(k)
					except:
						pass
	return vocabData

# Build class heirarchy recursively
def build_class(crmName, parent, vocabData):

	data = vocabData[crmName]
	name = str(data['className'])

	# check to see if we already exist
	# nb globals() here is only this module
	if name in globals():
		c = globals()[name]
		try:	
			c.__bases__ += (parent,)
		except:
			print("MRO FAILURE: %r --> %r + %r" % (c, c.__bases__, parent))
			raise
		return

	c = type(name, (parent,), {'__doc__': data['desc']})
	globals()[name] = c
	data['class'] = c
	if not ":" in crmName:
		c._type = "crm:%s" % crmName
	else:
		c._type = crmName
	c._uri_segment = name
	c._properties = {}
	c._all_properties = {}
	c._okayToUse = int(data['okay'])

	factory._all_classes[name] = c

	# Set up real properties
	for p in data['props']:
		name = p['name']
		if not ":" in name:
			name = "crm:%s" % name
		rng = p['range']
		ccname = p['propName']
		if p['inverse']:
			i = p['inverse']
			if i[0] == "P":
				invRdf = "crm:%s" % i
			else:
				invRdf = i
		else:
			invRdf = ""

		okay = p['okay']
		if not okay:
			okay = '1'
		okay = int(okay)
		mult = p['multiple']
		if not mult:
			mult = '0'
		mult = int(mult)

		# can't guarantee all classes have been built at this stage :(
		c._properties[ccname] = {"rdf": name, 
			"rangeStr": rng,
			"inverseRdf": invRdf,
			"okayToUse": okay,
			"multiple": mult}
 
	# Build subclasses
	for s in data['subs']:
		build_class(s, c, vocabData)

def build_classes(fn=None, topClass=None):
	# Default to building our core dataset

	if not fn:
		dd = os.path.join(os.path.dirname(__file__), 'data')
		fn = os.path.join(dd, 'crm_vocab.tsv')
		topClass = 'E1_CRM_Entity'

	vocabData = process_tsv(fn)

	# Everything can have an id, a type, a label, a description
	build_class(topClass, BaseResource, vocabData)

	# And reset definitions now we have class objects
	for v in vocabData.values():
		c = v['class']
		if c is BaseResource:
			continue
		for (name, value) in c._properties.items():
			# recreate with namedtuple
			rngs = value.get('rangeStr', None)
			if not rngs:
				# Precomputed ones like rdf:type
				continue

			inverse = None
			rngd = vocabData.get(value['rangeStr'], None)
			if rngs in ['rdfs:Literal', 'xsd:dateTime', 'xsd:string', 'rdfs:Class']:
					rng = str 
			elif not rngd:
				raise ConfigurationError("Failed to get range for %s property %s - %s" % (c, name, rngs))
			else:
				rng = rngd['class']
				# and add inverse prop name from range
				if rng == None:
					# Uh-oh!
					raise ConfigurationError("Class %s did not get built, range of %s" % (value['rangeStr'], name))
				for (ik, iv) in rng._properties.items():
					if iv['inverseRdf'] == value['rdf']:
						inverse = ik
						break

			c._all_properties[name] = PropInfo(name,
				value['rdf'], rng, inverse,
				value.get('inverseRdf', None),
				value.get('multiple', 1),
				value.get('okayToUse', 0)
			)

	# set all of the classhiers
	for v in vocabData.values():
		c = v['class']
		c._classhier = inspect.getmro(c)[:-1]
		try:
			del c._properties
		except:
			# Never had it set?
			pass

# XXX This should be invoked rather than inline so the module can be loaded
# and a different context used. But for now ...
# Build the factory first, so properties can be added to key_order
factory = CromulentFactory("http://lod.example.org/museum/", context="https://linked.art/ns/v1/linked-art.json")
build_classes()
# Need to then configure the boundary classes after they're created



# -*- coding: utf-8 -*-

import re
import warnings
from collections import namedtuple

from cromulent import model, vocab

#mark - Mapping Dictionaries

# TODO: can this be refactored somewhere?
CURRENCY_MAPPING = {
	'österreichische schilling': 'at shillings',
	'florins': 'de florins',
	'fl': 'de florins',
	'fl.': 'de florins',
	'pounds': 'gb pounds',
	'livres': 'fr livres',
	'guineas': 'gb guineas',
	'reichsmark': 'de reichsmarks'
}

#mark - Dimensions

NEXT_FINER_DIMENSION_UNIT = {
	'inches': None,
	'feet': 'inches',
	'cm': None,
	'fr_feet': 'fr_inches',
	'fr_inches': 'ligne'
}
NUMBER_PATTERN = r'((?:\d+\s+\d+/\d+)|(?:\d+/\d+)|(?:\d+(?:[.,]\d+)?))'
UNIT_PATTERN = r'''('|"|d(?:[.]?|uymen)|pouc[e.]s?|in(?:ch(?:es)?|[.]?)|'''\
				r'''pieds?|v[.]?|voeten|f(?:eet|oot|t[.]?)|cm|lignes?|linges?)'''
DIMENSION_PATTERN = '(%s\\s*(?:%s)?)' % (NUMBER_PATTERN, UNIT_PATTERN)
DIMENSION_RE = re.compile(r'\s*%s' % (DIMENSION_PATTERN,))

SIMPLE_WIDTH_HEIGHT_PATTERN = r'(?:\s*((?<!\w)[wh]|width|height))?'
SIMPLE_DIMENSIONS_PATTERN_X1 = ''\
	r'(?P<d1>(?:%s\s*)+)(?P<d1w>%s)' % (DIMENSION_PATTERN, SIMPLE_WIDTH_HEIGHT_PATTERN)
SIMPLE_DIMENSIONS_RE_X1 = re.compile(SIMPLE_DIMENSIONS_PATTERN_X1)
SIMPLE_DIMENSIONS_PATTERN_X2a = ''\
	r'(?P<d1>(?:%s\s*)+)(?P<d1w>%s)'\
	r'(?:,)?\s*(x|by)'\
	r'(?P<d2>(?:\s*%s)+)(?P<d2w>%s)' % (
		DIMENSION_PATTERN,
		SIMPLE_WIDTH_HEIGHT_PATTERN,
		DIMENSION_PATTERN,
		SIMPLE_WIDTH_HEIGHT_PATTERN)
SIMPLE_DIMENSIONS_PATTERN_X2b = ''\
	r'(?P<d1w>%s)\s*(?P<d1>(?:%s\s*)+)'\
	r'*(?:(?:,)?\s*(x|by)|,\s*)'\
	r'(?P<d2w>%s)\s*(?P<d2>(?:\s*%s)+)' % (
		SIMPLE_WIDTH_HEIGHT_PATTERN,
		DIMENSION_PATTERN,
		SIMPLE_WIDTH_HEIGHT_PATTERN,
		DIMENSION_PATTERN)
SIMPLE_DIMENSIONS_RE_X2a = re.compile(SIMPLE_DIMENSIONS_PATTERN_X2a)
SIMPLE_DIMENSIONS_RE_X2b = re.compile(SIMPLE_DIMENSIONS_PATTERN_X2b)

# Haut 14 pouces, large 10 pouces
FRENCH_DIMENSIONS_PATTERN = r'[Hh](?:(?:aut(?:eur|[.])?)|[.])\s*(?P<d1>(?:%s\s*)+),? '\
							r'[Ll](?:(?:arge?(?:ur|[.])?)|[.])\s*(?P<d2>(?:%s\s*)+)' % (
								DIMENSION_PATTERN, DIMENSION_PATTERN)
FRENCH_DIMENSIONS_RE = re.compile(FRENCH_DIMENSIONS_PATTERN)

# Hoog. 1 v. 6 d., Breed 2 v. 3 d.
DUTCH_DIMENSIONS_PATTERN = r'(?P<d1w>[Hh]oogh?[.]?|[Bb]reedt?) (?P<d1>(?:%s\s*)+), '\
							r'(?P<d2w>[Hh]oogh?[.]?|[Bb]reedt?) (?P<d2>(?:%s\s*)+)' % (
								DIMENSION_PATTERN, DIMENSION_PATTERN)
DUTCH_DIMENSIONS_RE = re.compile(DUTCH_DIMENSIONS_PATTERN)

Dimension = namedtuple("Dimension", [
	'value',	# numeric value
	'unit',		# unit
	'which'		# e.g. width, height, ...
])

def _canonical_value(value):
	try:
		value = value.replace(',', '.')
		parts = value.split(None, 1)
		if len(parts) > 1 and '/' in parts[1]:
			intpart = int(parts[0])
			numer, denom = map(int, parts[1].split('/', 1))
			fracpart = float(numer)/denom
			value = str(intpart + fracpart)
		elif len(parts) == 1 and '/' in parts[0]:
			intpart = 0
			numer, denom = map(int, parts[0].split('/', 1))
			fracpart = float(numer)/denom
			value = str(intpart + fracpart)
		if value.startswith('.'):
			value = '0' + value
		
		try:
			return int(value)
		except ValueError:
			return float(value)
	except:
		pass
	return None

def _canonical_unit(value):
	inches = {'duymen', 'd.', 'd', '"'}
	feet = {'pieds', 'pied', 'feet', 'foot', 'voeten', 'v.', 'v', "'"}
	fr_inches = {'pouces', 'pouce', 'pouc.'}
	fr_feet = {'pieds', 'pied'}
	if value is None:
		return None
	value = value.lower()
	if value in fr_inches:
		return 'fr_inches'
	if value in fr_feet:
		return 'fr_feet'
	if 'ligne' in value or 'linge' in value:
		return 'ligne'
	if 'in' in value or value in inches:
		return 'inches'
	if 'ft' in value or value in feet:
		return 'feet'
	if 'cm' in value:
		return 'cm'
	return None

def _canonical_which(value):
	if not value:
		return None
	value = value.strip().lower()
	if value.startswith('w'):
		return 'width'
	if value.startswith('h'):
		return 'height'
	warnings.warn('*** Unknown which dimension: %s' % (value,))
	return None

def parse_simple_dimensions(value, which=None):
	'''
	Parse the supplied string for dimensions (value + unit), and return a list of
	`Dimension`s, optionally setting the `which` property to the supplied value.

	Examples:

	1 cm
	2ft
	5 pieds
	'''
	if value is None:
		return None
	value = value.strip()
	dims = []
# 	warnings.warn('DIMENSION: %s' % (value,))

	last_unit = None
	for match in re.finditer(DIMENSION_RE, value):
		# warnings.warn('--> match %s' % (match,))
		matched_value = _canonical_value(match.group(2))
		if not matched_value:
			warnings.warn('*** failed to canonicalize dimension value: %s' % (match.group(2),))
			return None
		unit_value = match.group(3)
		matched_unit = _canonical_unit(unit_value)
		if matched_unit is None:
			matched_unit = NEXT_FINER_DIMENSION_UNIT.get(last_unit)
		if unit_value and not matched_unit:
			warnings.warn('*** not a recognized unit: %s' % (unit_value,))
		which = _canonical_which(which)
		dim = Dimension(value=matched_value, unit=matched_unit, which=which)
		last_unit = matched_unit
		dims.append(dim)
	if not dims:
		return None
	return dims

def normalized_dimension_object(dimensions, source=None, **kwargs):
	'''
	Normalizes the given `dimensions`, or returns `None` is normalization fails.

	Returns a tuple of the normalized data, and a label which preserves the original
	set of dimensions.

	For example, the input:

		[
			Dimension(value=10, unit='feet', which=None),
			Dimension(value=3, unit='inches', which=None),
		]

	results in the output:

		(
			Dimension(value=123.0, unit='inches', which=None),
			"10 feet, 3 inches"
		)
	'''
	normalized = normalize_dimension(dimensions, source=source, **kwargs)
	if not normalized:
		return None
	labels = []
	for dim in dimensions:
		if dim.unit == 'inches':
			units = ('inch', 'inches')
		elif dim.unit == 'feet':
			units = ('foot', 'feet')
		elif dim.unit == 'fr_feet':
			units = ('French foot', 'French feet')
		elif dim.unit == 'fr_inches':
			units = ('French inch', 'French inches')
		elif dim.unit == 'cm':
			units = ('cm', 'cm')
		elif dim.unit == 'ligne':
			units = ('ligne', 'lignes')
		elif dim.unit is None:
			units = ('', '')
		else:
			warnings.warn('*** unrecognized unit: {dim.unit}')
			return None
		unit = units[0] if (float(dim.value) == 1.0) else units[1]
		if unit:
			label = '%s %s' % (dim.value, unit)
		else:
			label = str(dim.value)
		labels.append(label)
	label = ', '.join(labels)
	return normalized, label

def normalize_dimension(dimensions, source=None):
	'''
	Given a list of `Dimension`s, normalize them into a single Dimension (e.g. values in
	both feet and inches become a single dimension of inches).

	If the values cannot be sensibly combined (e.g. inches + centimeters), returns `None`.
	'''
	unknown = 0
	inches = 0
	fr_inches = 0
	centimeters = 0
	used_unknown = False
	used_inches = False
	used_fr_inches = False
	used_centimeters = False
	which = None
	for dim in dimensions:
		which = dim.which
		if dim.unit == 'inches':
			inches += dim.value
			used_inches = True
		elif dim.unit == 'feet':
			inches += 12 * dim.value
			used_inches = True
		elif dim.unit == 'cm':
			centimeters += dim.value
			used_centimeters = True
		elif dim.unit == 'fr_feet':
			fr_inches += 12.0 * dim.value
			used_fr_inches = True
		elif dim.unit == 'fr_inches':
			fr_inches += dim.value
			used_fr_inches = True
		elif dim.unit == 'ligne':
			fr_inches += dim.value / 12.0
			used_fr_inches = True
		elif dim.unit is None:
			unknown += dim.value
			used_unknown = True
		else:
			warnings.warn('*** unrecognized unit: %s' % (dim.unit,))
			return None

	used_systems = 0
	for used in (used_inches, used_fr_inches, used_centimeters, used_unknown):
		if used:
			used_systems += 1
	if used_systems != 1:
		if source:
			warnings.warn('*** dimension used a mix of metric, imperial, and/or unknown: '\
							'%r; source is %r' % (dimensions, source))
		else:
			warnings.warn('*** dimension used a mix of metric, imperial, and/or unknown: '\
							'%r' % (dimensions,))
		return None
	if fr_inches:
		return Dimension(value=fr_inches, unit='fr_inches', which=which)
	if inches:
		return Dimension(value=inches, unit='inches', which=which)
	if centimeters:
		return Dimension(value=centimeters, unit='cm', which=which)
	return Dimension(value=unknown, unit=None, which=which)

def extract_physical_dimensions(dimstr, **kwargs):
	dimensions = dimensions_cleaner(dimstr, **kwargs)
	if dimensions:
		for orig_d in dimensions:
			dimdata = normalized_dimension_object(orig_d, source=dimstr)
			if dimdata:
				dimension, label = dimdata
				if dimension.which == 'height':
					dim = vocab.Height(ident='')
				elif dimension.which == 'width':
					dim = vocab.Width(ident='')
				else:
					dim = vocab.PhysicalDimension(ident='')
				dim.value = dimension.value
				dim.identified_by = model.Name(ident='', content=label)
				unit = vocab.instances.get(dimension.unit)
				if unit:
					dim.unit = unit
				yield dim

def dimensions_cleaner(value, default_unit=None):
	'''
	Attempt to parse a set of dimensions from the given string.

	Returns a tuple of `pipeline.util.Dimension` objects if parsing succeeds,
	None otherwise.
	'''
	if value is None:
		return None
	cleaners = [
		simple_dimensions_cleaner_x2,
		french_dimensions_cleaner_x2,
		dutch_dimensions_cleaner_x2,
		simple_dimensions_cleaner_x1
	]
	for cleaner in cleaners:
		dimensions = cleaner(value)
		if dimensions:
			if default_unit:
				for i, d_list in enumerate(dimensions):
					for j, d in enumerate(d_list):
						if d.unit is None:
							dimensions[i][j] = Dimension(value=d.value, unit=default_unit, which=d.which)
			return dimensions
	return None

def french_dimensions_cleaner_x2(value):
	'''Attempt to parse 2 dimensions from a French-formatted string.'''
	# Haut 14 pouces, large 10 pouces

	match = FRENCH_DIMENSIONS_RE.match(value)
	if match:
		groups = match.groupdict()
		dim1 = parse_simple_dimensions(groups['d1'], 'h')
		dim2 = parse_simple_dimensions(groups['d2'], 'w')
		if dim1 and dim2:
			return (dim1, dim2)
		warnings.warn('dim1: %s %s h' % (dim1, groups['d1']))
		warnings.warn('dim2: %s %s w' % (dim2, groups['d2']))
		warnings.warn('*** Failed to parse dimensions: %s' % (value,))
	return None

def dutch_dimensions_cleaner_x2(value):
	'''Attempt to parse 2 dimensions from a Dutch-formatted string.'''
	# Hoog. 1 v. 6 d., Breed 2 v. 3 d.
	# Breedt 6 v., hoog 3 v

	match = DUTCH_DIMENSIONS_RE.match(value)
	if match:
		groups = match.groupdict()
		height = 'h'
		width = 'w'
		if 'breed' in groups['d1w'].lower():
			height, width = width, height

		dim1 = parse_simple_dimensions(groups['d1'], height)
		dim2 = parse_simple_dimensions(groups['d2'], width)
		if dim1 and dim2:
			return (dim1, dim2)
		warnings.warn('dim1: %s %s h' % (dim1, groups['d1']))
		warnings.warn('dim2: %s %s w' % (dim2, groups['d2']))
		warnings.warn('*** Failed to parse dimensions: %s' % (value,))
	return None

def simple_dimensions_cleaner_x1(value):
	'''Attempt to parse 1 dimension from a string.'''
	# 1 cm
	# 1' 2"
	# 1 ft. 2 in. h

	match = SIMPLE_DIMENSIONS_RE_X1.match(value)
	if match:
		groups = match.groupdict()
		dim1 = parse_simple_dimensions(groups['d1'], groups['d1w'])
		if dim1:
			return (dim1,)
	return None

def simple_dimensions_cleaner_x2(value):
	'''Attempt to parse 2 dimensions from a string.'''
	# 1 cm x 2 in
	# 1' 2" by 3 cm
	# 1 ft. 2 in. h by 3 cm w

	for pattern in (SIMPLE_DIMENSIONS_RE_X2a, SIMPLE_DIMENSIONS_RE_X2b):
		match = pattern.match(value)
		if match:
			groups = match.groupdict()
			dim1 = parse_simple_dimensions(groups['d1'], groups['d1w'])
			dim2 = parse_simple_dimensions(groups['d2'], groups['d2w'])
			if dim1 and dim2:
				return (dim1, dim2)
			warnings.warn('dim1: %s %s %s' % (dim1, groups['d1'], groups['d1w']))
			warnings.warn('dim2: %s %s %s' % (dim2, groups['d2'], groups['d2w']))
			warnings.warn('*** Failed to parse dimensions: %s' % (value,))
	return None

#mark - Monetary Values

def extract_monetary_amount(data, add_citations=False, currency_mapping=CURRENCY_MAPPING, source_mapping=None):
	'''
	Returns a `MonetaryAmount`, `StartingPrice`, or `EstimatedPrice` object
	based on properties of the supplied `data` dict. If no amount or currency
	data is found in found, returns `None`.

	For `EstimatedPrice`, values will be accessed from these keys:
		- amount: `est_price_amount` or `est_price`
		- currency: `est_price_currency` or `est_price_curr`
		- note: `est_price_note` or `est_price_desc`
		- bibliographic statement: `est_price_citation`

	For `StartingPrice`, values will be accessed from these keys:
		- amount: `start_price_amount` or `start_price`
		- currency: `start_price_currency` or `start_price_curr`
		- note: `start_price_note` or `start_price_desc`
		- bibliographic statement: `start_price_citation`

	For `MonetaryAmount` prices, values will be accessed from these keys:
		- amount: `price_amount` or `price`
		- currency: `price_currency` or `price_curr`
		- note: `price_note` or `price_desc`
		- bibliographic statement: `price_citation`
	'''
	amount_type = 'Price'
	if 'price' in data or 'price_amount' in data or 'amount' in data:
		amnt = model.MonetaryAmount(ident='')
		price_amount = data.get('price_amount', data.get('price', data.get('amount')))
		price_currency = data.get('currency', data.get('price_currency', data.get('price_curr')))
		note = data.get('price_note', data.get('price_desc', data.get('note')))
		cite = data.get('price_citation', data.get('citation'))
		source = data.get('price_source', '')
	elif 'est_price' in data or 'est_price_amount' in data:
		amnt = vocab.EstimatedPrice(ident='')
		price_amount = data.get('est_price_amount', data.get('est_price'))
		price_currency = data.get('currency', data.get('est_price_currency', data.get('est_price_curr')))
		amount_type = 'Estimated Price'
		note = data.get('est_price_note', data.get('est_price_desc', data.get('note')))
		cite = data.get('est_price_citation', data.get('citation'))
		source = data.get('est_price_source', data.get('est_price_so', ''))
	elif 'start_price' in data or 'start_price_amount' in data:
		amnt = vocab.StartingPrice(ident='')
		price_amount = data.get('start_price_amount', data.get('start_price'))
		price_currency = data.get('currency', data.get('start_price_currency', data.get('start_price_curr')))
		amount_type = 'Starting Price'
		note = data.get('start_price_note', data.get('start_price_desc', data.get('note')))
		cite = data.get('start_price_citation', data.get('citation'))
		source = data.get('start_price_source', data.get('start_price_so', ''))
	else:
		return None

	price_amount_label = price_amount
	if price_amount or price_currency:
		if cite and add_citations:
			amnt.referred_to_by = vocab.BibliographyStatement(ident='', content=cite)
		if note:
			amnt.referred_to_by = vocab.Note(ident='', content=note)

		if price_amount:
			try:
				value = price_amount
				value = value.replace('[?]', '')
				value = value.replace('?', '')
				value = value.strip()
				if re.search(re.compile(r',\d\d\d'), value):
					value = value.replace(',', '')
				value = float(value)
				price_amount_label = '{:,}'.format(round(value, 3))
				amnt.value = value
			except ValueError:
				amnt._label = price_amount
				amnt.identified_by = model.Name(ident='', content=price_amount)
	# 			warnings.warn(f'*** Not a numeric price amount: {value}')
		if price_currency:
			price_currency_key = price_currency
			try:
				price_currency_key = currency_mapping[price_currency_key.lower()]
			except KeyError:
				pass
			if isinstance(price_currency_key, model.BaseResource):
				amnt.currency = price_currency_key
			elif price_currency_key in vocab.instances:
				amnt.currency = vocab.instances[price_currency_key]
			else:
				warnings.warn('*** No currency instance defined for %s' % (price_currency_key,))
		if price_amount_label and price_currency:
			amnt._label = '%s %s' % (price_amount_label, price_currency)
		elif price_amount:
			amnt._label = '%s' % (price_amount,)
		return amnt
	return None




# This assumes the default CIDOC-CRM, even though the model code 
# can generate classes for any ontology

from .model import Identifier, Mark, HumanMadeObject, Type, \
	Person, Material, MeasurementUnit, Place, Dimension, Currency, \
	ConceptualObject, TimeSpan, Actor, PhysicalThing, Language, \
	LinguisticObject, InformationObject, Formation, Dissolution, \
	Activity, Group, Name, MonetaryAmount, Right, Encounter, \
	Destruction, AttributeAssignment, BaseResource, PhysicalObject, \
	Acquisition, HumanMadeFeature, VisualItem, Set, Birth, Death, \
	PropositionalObject, Payment, Creation, Phase, Period, \
	Production, Event, DigitalObject, TransferOfCustody, \
	STR_TYPES, factory

# Add classified_as initialization hack for all resources
def post_init(self, **kw):
	if self.__class__._classification:
		for t in self._classification:
			self.classified_as = t
BaseResource._post_init = post_init

instances = {}
instance_types = {}

def register_vocab_class(name, data):
	parent = data['parent']
	id = data['id']
	label = data['label']
	vocab = data.get('vocab', 'aat')

	c = type(name, (parent,), {})
	if id.startswith('http'):
		t = Type(id)
	else:
		t = Type("http://vocab.getty.edu/%s/%s" % (vocab, id))
	t._label = label
	instance_types[name] = t
	c._classification = [t]		
	if "metatype" in data:
		t.classified_as = instances[data['metatype']]
	c._type = None # To avoid conflicting with parent class
	globals()[name] = c	
	return c

def register_aat_class(name, data):
	data['vocab'] = 'aat'
	register_vocab_class(name, data)

def register_instance(name, data):
	parent = data['parent']
	id = data['id']
	vocab = data.get('vocab', 'aat')
	label = data['label']

	if id.startswith('http'):
		t = parent(id)
	else:
		t = parent("http://vocab.getty.edu/%s/%s" % (vocab, id))
	t._label = label
	instances[name] = t
	return t

# Meta meta
ext_classes = {

	"LocalNumber": {"parent":Identifier, "id":"300404621", "label": "Owner-Assigned Number"},	
	"SystemNumber": {"parent":Identifier, "id":"300435704", "label": "System-Assigned Number"},
	"AccessionNumber": {"parent":Identifier, "id":"300312355", "label": "Accession Number"},
	"LotNumber": {"parent": Identifier, "id": "300404628", "label": "Lot Number"},
	"IsbnIdentifier": {"parent": Identifier, "id": "300417443", "label": "ISBN Identifier"},
	"IssnIdentifier": {"parent": Identifier, "id": "300417430", "label": "ISSN Identifier"},
	"CodenIdentifier": {"parent": Identifier, "id": "300417431", "label": "CODEN Identifer"},
	"DoiIdentifier": {"parent": Identifier, "id": "300417432", "label": "DOI Identifer"},
	"VolumeNumber": {"parent": Identifier, "id": "300265632", "label": "Volume"},
	"IssueNumber": {"parent": Identifier, "id": "300312349", "label": "Issue"},

	"EmailAddress": {"parent": Identifier, "id":"300435686", "label": "Email Address"},
	"StreetAddress": {"parent": Identifier, "id":"300435687", "label": "Street Address"},
	"TelephoneNumber": {"parent": Identifier, "id": "300435688", "label": "Telephone Number"},
	"FaxNumber": {"parent": Identifier, "id": "300435689", "label": "Fax Number"},
	"StreetNumber": {"parent": Identifier, "id":"300419272", "label": "Street Number"},
	"StreetName": {"parent": Identifier, "id": "300419273", "label": "Street Name"},
	"PostalCode": {"parent": Identifier, "id": "300419274", "label": "Postal Code"},

	"DigitalImage": {"parent": DigitalObject, "id": "300215302", "label": "Digital Image"},
	"WebPage": {"parent": DigitalObject, "id":"300264578", "label": "Web Page"},

	"OwnershipRight": {"parent": Right, "id":"300055603", "label": "Ownership Right"},
	"CustodyRight": {"parent": Right, "id":"300411616", "label": "Custody Right"},
	"CopyrightRight": {"parent": Right, "id":"300055598", "label": "Copyright"},

	"OwnershipPhase": {"parent": Phase, "id": "300055603", "label": "Ownership Phase"},

	"_BriefText": {"parent": LinguisticObject, "id": "300418049", "label":"Brief Text"},
	"Note": {"parent": LinguisticObject, "id":"300027200", "label": "Note", "metatype": "brief text"},
	"Description": {"parent": LinguisticObject, "id":"300411780", "label": "Description", "metatype": "brief text"},
	"Abstract": {"parent": LinguisticObject, "id":"300026032", "label": "Abstract", "metatype": "brief text"},
	"Citation": {"parent": LinguisticObject, "id":"300311705", "label": "Citation", "metatype": "brief text"},

	"CreditStatement": {"parent": LinguisticObject, "id": "300026687", "label": "Credit Statement", "metatype": "brief text"},
	"RightsStatement": {"parent": LinguisticObject, "id": "300435434", "label": "Copyright/License Statement", "metatype": "brief text"},
	"ValuationStatement": {"parent": LinguisticObject, "id": "300435426", "label": "Appraisal Statement", "metatype": "brief text"},
	"ConditionStatement": {"parent": LinguisticObject, "id": "300435425", "label": "Condition Statement", "metatype": "brief text"},
	"SignificanceStatement": {"parent": LinguisticObject, "id": "300435427", "label": "Significance Statement", "metatype": "brief text"},
	"ContextStatement": {"parent": LinguisticObject, "id": "300435428", "label": "Historical/Cultural Context", "metatype": "brief text"},
	"SourceStatement": {"parent": LinguisticObject, "id": "300404764", "label": "Source Statement", "metatype": "brief text"},
	"PropertyStatusStatement": {"parent": LinguisticObject, "id": "300438433", "label": "Property Status Statement", "metatype": "brief text"},

	"MarkingsStatement": {"parent": LinguisticObject, "id": "300435420", "label": "Markings", "metatype": "brief text"},
	"InscriptionStatement": {"parent": LinguisticObject, "id": "300435414", "label": "Inscription", "metatype": "brief text"},
	"SignatureStatement": {"parent": LinguisticObject, "id": "300435415", "label": "Signature", "metatype": "brief text"},
	"WatermarksStatement": {"parent": LinguisticObject, "id": "300435421", "label": "Watermarks", "metatype": "brief text"},
	"MaterialStatement": {"parent": LinguisticObject, "id": "300435429", "label": "Material Statement", "metatype": "brief text"},
	"PhysicalStatement": {"parent": LinguisticObject, "id": "300435452", "label": "Physical Statement", "metatype": "brief text"},
	"DimensionStatement": {"parent": LinguisticObject, "id": "300435430", "label": "Dimension Statement", "metatype": "brief text"},
	"CultureStatement": {"parent": LinguisticObject, "id": "300435431", "label": "Culture Statement", "metatype": "brief text"},
	"PeriodStatement": {"parent": LinguisticObject, "id": "300435432", "label": "Period Statement", "metatype": "brief text"},

	"ProductionStatement": {"parent": LinguisticObject, "id": "300435436", "label": "Production Statement", "metatype": "brief text" },
	"AcquisitionStatement": {"parent": LinguisticObject, "id": "300435439", "label": "Acquisition Statement", "metatype": "brief text"},
	"ProvenanceStatement": {"parent": LinguisticObject, "id":"300435438", "label": "Provenance Statement", "metatype": "brief text"},
	"BibliographyStatement": {"parent": LinguisticObject, "id": "300026497", "label": "Bibliography Statement", "metatype": "brief text"},
	"ExhibitionStatement": {"parent": LinguisticObject, "id":"300435424", "label":"Exhibition Statement", "metatype": "brief text"},
	"PriceStatement": {"parent": LinguisticObject, "id":"300055694", "label":"Price Statement", "metatype": "brief text"},

	"BiographyStatement": {"parent": LinguisticObject, "id":"300435422", "label": "Biography Statement", "metatype": "brief text"},
	"EditionStatement": {"parent": LinguisticObject, "id":"300435435", "label": "Edition Statement", "metatype": "brief text"},
	"PaginationStatement": {"parent": LinguisticObject, "id":"300435440", "label": "Pagination Statement", "metatype": "brief text"},
	"FoliationStatement": {"parent": LinguisticObject, "id":"300435441", "label": "Foliation Statement", "metatype": "brief text"},
	"LanguageStatement": {"parent": LinguisticObject, "id": "300435433", "label": "Language Statement", "metatype": "brief text"},

	"TranscribedInscription": {"parent": LinguisticObject, "id": "300028702", "label": "Transcribed Inscription", "metatype": "brief text"},
	"TranscribedSignature": {"parent": LinguisticObject, "id": "300028705", "label": "Transcribed Signature", "metatype": "brief text"},

	"CatalogueRaisonneText": {"parent": LinguisticObject, "id":"300026061", "label": "Catalogue Raisonne", "metatype": "work type"},
	"AuctionCatalogText": {"parent": LinguisticObject, "id":"300026068", "label": "Auction Catalog", "metatype": "work type"},
	"SalesCatalogText": {"parent": LinguisticObject, "id": "300026074", "label": "Sales Catalog", "metatype": "work type"},
	"LotteryCatalogText": {"parent": LinguisticObject, "id":"300438603", "label": "Lottery Catalog", "metatype": "work type"},
	"ExhibitionCatalogText": {"parent": LinguisticObject, "id": "300026096", "label": "Exhibition Catalog", "metatype": "work type"},
	"AccessionCatalogText": {"parent": LinguisticObject, "id": "300026617", "label": "Accession Catalog", "metatype": "work type"},
	"AccountBookText": {"parent": LinguisticObject, "id": "300027483", "label": "Account Book", "metatype": "work type"},
	"RegisterText": {"parent": LinguisticObject, "id":"300027168", "label": "Register", "metatype": "work type"},
	"PageText": {"parent": LinguisticObject, "id":"300194222", "label": "Page", "metatype": "work type"},
	"FolioText": {"parent": LinguisticObject, "id":"300189604", "label": "Folio", "metatype": "work type"},
	"ParagraphText": {"parent": LinguisticObject, "id":"300417223", "label": "Paragraph"},
	"Annotation": {"parent": LinguisticObject, "id":"300026100", "label":"Annotation"},

	"DataRecord": {"parent": LinguisticObject, "id":"300438434", "label": "Entry or Record"},
	"Heading": {"parent": LinguisticObject, "id": "300200862", "label": "Heading"},
	"SubHeading": {"parent": LinguisticObject, "id": "300419992", "label": "SubHeading"},

	"JournalText": {"parent": LinguisticObject, "id":"300215390", "label": "Journal"},
	"IssueText": {"parent": LinguisticObject, "id":"300312349", "label": "Issue"},
	"VolumeText": {"parent": LinguisticObject, "id":"300265632", "label": "Volume"},
	"ArticleText": {"parent": LinguisticObject, "id":"300048715", "label": "Article"},
	"ChapterText": {"parent": LinguisticObject, "id":"300311699", "label": "Chapter"},
	"SeriesText": {"parent": LinguisticObject, "id":"300027349", "label": "Series"},
	"ProceedingsText": {"parent": LinguisticObject, "id":"300027316", "label": "Proceedings"},
	"MonographText": {"parent": LinguisticObject, "id":"300060417", "label": "Monograph"},
	"PatentText": {"parent": LinguisticObject, "id":"300027832", "label": "Patent"},
	"ThesisText": {"parent": LinguisticObject, "id":"300028028", "label": "Thesis"},
	"TechnicalReportText": {"parent": LinguisticObject, "id":"300027323", "label": "Technical Report"},
	"DocManuscriptText": {"parent": LinguisticObject, "id":"300028579", "label": "Document Manuscript"},
	"AudioVisualContent": {"parent": LinguisticObject, "id":"300028045", "label": "A/V Content"},

	"Height":     {"parent": Dimension, "id":"300055644", "label": "Height"},
	"Width":      {"parent": Dimension, "id":"300055647", "label": "Width"},
	"Depth":      {"parent": Dimension, "id":"300072633", "label": "Depth"},
	"Diameter":   {"parent": Dimension, "id":"300055624", "label": "Diameter"},
	"Weight":     {"parent": Dimension, "id":"300056240", "label": "Weight"},
	"Color":      {"parent": Dimension, "id":"300080438", "label": "Color"}, # individual colors as dimensions, 56130 is concept of color
	"SequencePosition": {"parent": Dimension, "id":"300010269", "label":"Sequence Position"},
	"PhysicalDimension": {"parent": Dimension, "id":"300055642", "label":"Unknown physical dimension"},

	"Gallery":      {"parent": Place, "id":"300240057", "label": "Gallery (place)"},
	"AuctionHouse": {"parent": Place, "id":"300005234", "label": "Auction House (place)"},
	"MuseumPlace":  {"parent": Place, "id":"300005768", "label": "Museum (place)"},
	"ExhibitionPlace": {"parent": Place, "id":"300005748", "label": "Exhibition (place)"},
	"CityBlock":    {"parent": Place, "id":"300008077", "label": "City Block"},
	"City":         {"parent": Place, "id":"300008389", "label": "City"},
	"Province":     {"parent": Place, "id":"300000774", "label": "Province"},
	"Nation":       {"parent": Place, "id":"300128207", "label": "Nation"},
	"Sovereign":	{"parent": Place, "id":"300232420", "label": "Sovereign State"},
	"StoragePlace": {"parent": Place, "id":"300150151", "label": "Storage (place)"},

	"Building":     {"parent": HumanMadeObject, "id": "300004792", "label":"Building"},

	"AuctionHouseOrg": {"parent": Group, "id": "300417515", "label": "Auction House (organization)"},
	"MuseumOrg":   {"parent": Group, "id":"300312281", "label": "Museum"},
	"Department":  {"parent": Group, "id":"300263534", "label": "Department"},
	"School":      {"parent": Group, "id":"300404284", "label": "School"},
	"Studio":      {"parent": Group, "id":"300404275", "label": "Studio"},
	"Circle":      {"parent": Group, "id":"300404283", "label": "Circle"},
	"Workshop":    {"parent": Group, "id":"300404274", "label": "Workshop"},
	"Atelier":     {"parent": Group, "id":"300404277", "label": "Atelier"},
	"FollowerGroup": {"parent": Group, "id":"300404282", "label": "Follower"},
	"PupilGroup":  {"parent": Group, "id":"300404279", "label": "Pupil"},

	"Nationality": {"parent": Type, "id":"300379842", "label": "Nationality"},
	"Gender":      {"parent": Type, "id":"300055147", "label": "Gender"},
	"Occupation":  {"parent": Type, "id":"300263369", "label": "Occupation"},
	"Shape":       {"parent": Type, "id":"300056273", "label": "Shape"},
	"Style": 	   {"parent": Type, "id":"300015646", "label": "Style"},

	"Auctioneer":  {"parent": Person, "id":"300025208", "label": "Auctioneer"}, # is this useful?

	"AuctionEvent": {"parent": Activity, "id":"300054751", "label": "Auction Event"},
	"Auction":     {"parent": Activity, "id":"300420001", "label": "Auction of Lot"}, # Individual auction-of-lot
	"LotteryDrawing": {"parent": Activity, "id":"300438601", "label": "Lottery Drawing"},
	"Lottery":     {"parent": Activity, "id":"300184191", "label": "Lottery"},

	"Negotiating": {"parent": Activity, "id":"300438558", "label": "Negotiating"},
	"Bidding":     {"parent": Creation, "id":"300054602", "label": "Bidding"}, # individual bid
	"Curating":    {"parent": Activity, "id":"300054277", "label": "Curating"},
	"Inventorying": {"parent": Encounter, "id":"300077506", "label": "Inventorying"},
	"Exhibition":  {"parent": Activity, "id":"300054766", "label": "Exhibiting"},
	"MultiExhibition": {"parent": Activity, "id":"300054773", "label": "Exhibiting in multiple locations"},
	"Publishing":  {"parent": Activity, "id":"300054686", "label": "Publishing"},
	"Distributing":  {"parent": Activity, "id": "300137595", "label": "Distributing"},
	"Hiring":      {"parent": Activity, "id":"300109703", "label": "Hiring"},
	"Expert":      {"parent": Activity, "id":"300417631", "label": "Expert"},
	"CommissairePriseur":  {"parent": Activity, "id":"300412173", "label": "Commissaire-priseur"},

	"Active":      {"parent": Activity, "id":"300393177", "label": "Professional Activities"},
	"Collecting": {"parent": Activity, "id": "300077121", "label": "Collecting"},
	"Creating": {"parent": Activity, "id": "300404387", "label": "Creating Artwork"},
	"Dealing": {"parent": Activity, "id": "300055675", "label": "Commercial Dealing in Artwork"},
	"Owning": {"parent": Activity, "id": "300055603", "label": "Owning"},
	"Promise": {"parent": Activity, "id": "300435599", "label": "Promise"},

	"Purchase":  {"parent":Acquisition, "id":"300077989", "label": "Purchasing"},
	"Assembling": {"parent": Activity, "id":"300077121", "label":"Assembling"},
	"Managing": {"parent": Activity, "id":"300054277", "label": "Managing"},
	"Storing": {"parent": Activity, "id":"300056390", "label": "Storing"},
	"Producing": {"parent": Activity, "id":"300054713", "label": "Producing"},

	"ProvenanceEntry": {"parent": Activity, "id":"300055863", "label": "Provenance Entry"},
	"ReturnOfLoan": {"parent": TransferOfCustody, "id":"300438467", "label": "Return of Loan"},
	"Theft": {"parent": TransferOfCustody, "id": "300055292", "label": "Theft"},
	"Looting": {"parent": TransferOfCustody, "id":"300379554", "label": "Looting"},
	"Loss": {"parent": TransferOfCustody, "id":"300417655", "label": "Loss"},
	"Loan": {"parent": TransferOfCustody, "id":"300417645", "label": "Loan"},
	"LongtermLoan": {"parent": TransferOfCustody, "id":"300417646", "label": "Long-term Loan"},
	"SaleOfStolenGoods": {"parent": TransferOfCustody, "id":"xxx", "label":"Sale of Stolen Goods"},

	"PossibleAssignment": {"parent": AttributeAssignment, "id":"300435722", "label": "Possibly"}, 
	"ProbableAssignment": {"parent": AttributeAssignment, "id":"300435721", "label": "Probably"},
	"ObsoleteAssignment": {"parent": AttributeAssignment, "id":"300404908", "label": "Obsolete"},

	"Measuring": {"parent": AttributeAssignment, "id":"300053578", "label":"Measuring"},
	"Identifying": {"parent": AttributeAssignment, "id":"300137570","label":"Identifying"},
	"Naming": {"parent": AttributeAssignment, "id":"300411672", "label":"Naming"},

	"ExhibitionIdea": {"parent": PropositionalObject, "id":"300417531", "label": "Exhibition"},

	"AuctionLotSet": {"parent": Set, "id":"300411307", "label": "Auction Lot"},
	"CollectionSet": {"parent": Set, "id":"300025976", "label": "Collection"},
	"ArchiveSet": {"parent": Set, "id":"300375748", "label": "Archive"}, # Maybe 300189759?
	"ArchiveGroupSet": {"parent": Set, "id":"300404022", "label": "Archival Grouping"},
	"ArchiveSubGroupSet": {"parent": Set, "id":"300404023", "label": "Archival SubGrouping"},
	
	"PrimaryName": {"parent": Name, "id":"300404670", "label": "Primary Name"},
	"SortName": {"parent": Name, "id":"300404672", "label": "Sorting Name"},

	"Title": {"parent": Name, "id":"300195168", "label": "Title"},
	"Subtitle": {"parent": Name, "id":"300312006", "label":"Subtitle"},
	"DisplayName": {"parent": Name, "id": "300404669", "label": "Display Title"},
	"TranslatedTitle": {"parent": Name, "id":"300417194", "label": "Translated Title"},

	"PersonalName": {"parent": Name, "id":"300266386", "label": "Personal Name"},
	"AliasName": {"parent": Name, "id": "300404664", "label": "Alias"},
	"MaidenName": {"parent": Name, "id": "300404682", "label": "Maiden Name"},
	"FormerName": {"parent": Name, "id": "300435719", "label": "Former Name"},
	"GivenName": {"parent": Name, "id":"300404651", "label": "Given Name"},
	"FamilyName": {"parent": Name, "id":"300404652", "label": "Family Name"},
	"MiddleName": {"parent": Name, "id":"300404654", "label": "Middle Name"},
	"NameSuffix": {"parent": Name, "id":"300404662", "label": "Name Suffix"},
	"NamePrefix": {"parent": Name, "id":"300404845", "label": "Name Prefix"},

	"CityName": {"parent": Name, "id": "300008389", "label": "City"},
	"CountryName": {"parent": Name, "id": "300128207", "label": "Country"},

	"Painting": {"parent": HumanMadeObject, "id": "300033618", "label": "Painting", "metatype": "work type"},
	"Sculpture": {"parent": HumanMadeObject, "id": "300047090", "label": "Sculpture", "metatype": "work type"},
	"Drawing": {"parent": HumanMadeObject, "id": "300033973", "label": "Drawing", "metatype": "work type"},
	"Miniature": {"parent": HumanMadeObject, "id": "300033936", "label": "Miniature", "metatype": "work type"},
	"Tapestry": {"parent": HumanMadeObject, "id": "300205002", "label": "Tapestry", "metatype": "work type"},
	"Furniture": {"parent": HumanMadeObject, "id": "300037680", "label": "Furniture", "metatype": "work type"},
	"Mosaic": {"parent": HumanMadeObject, "id": "300015342", "label": "Mosaic", "metatype": "work type"},
	"Photograph": {"parent": HumanMadeObject, "id": "300046300", "label": "Photograph", "metatype": "work type"},
	"Coin": {"parent": HumanMadeObject, "id": "300037222", "label": "Coin", "metatype": "work type"},
	"Vessel": {"parent": HumanMadeObject, "id": "300193015", "label": "Vessel", "metatype": "work type"},
	"Graphic": {"parent": HumanMadeObject, "id": "300015387", "label": "Graphic Art", "metatype": "work type"},
	"Enamel": {"parent": HumanMadeObject, "id": "300178264", "label": "Enamel", "metatype": "work type"},
	"Embroidery": {"parent": HumanMadeObject, "id": "300264024", "label": "Embroidery", "metatype": "work type"},
	"PhotographPrint": {"parent": HumanMadeObject, "id": "300127104", "label": "Photograph Print", "metatype": "work type"},
	"PhotographAlbum": {"parent": HumanMadeObject, "id": "300026695", "label": "Photograph Album", "metatype": "work type"},
	"PhotographBook": {"parent": HumanMadeObject, "id": "300265728", "label": "Photograph Book", "metatype": "work type"},
	"PhotographColor": {"parent": HumanMadeObject, "id": "300128347", "label": "Color Photograph", "metatype": "work type"},
	"PhotographBW": {"parent": HumanMadeObject, "id": "300128359", "label": "Black and White Photograph", "metatype": "work type"},
	"Negative": {"parent": HumanMadeObject, "id": "300127173", "label": "Photographic Negative", "metatype": "work type"},
	"Map": {"parent": HumanMadeObject, "id": "300028094", "label": "Map", "metatype": "work type"},
	"Clothing": {"parent": HumanMadeObject, "id": "300266639", 'label': "Clothing", "metatype": "work type"},
	"Furniture": {"parent": HumanMadeObject, "id":"300037680", "label": "Furniture", "metatype": "work type"},

	"Architecture": {"parent": HumanMadeObject, "id":"300263552", "label": "Architecture", "metatype": "work type"},
	"Armor": {"parent": HumanMadeObject, "id":"300036745", "label": "Armor", "metatype": "work type"},
	"Book": {"parent": HumanMadeObject, "id":"300028051", "label": "Book", "metatype": "work type"},
	"DecArts": {"parent": HumanMadeObject, "id":"300054168", "label": "Decorative Arts", "metatype": "work type"},
	"Implement": {"parent": HumanMadeObject, "id":"300024841", "label": "Implement", "metatype": "work type"},
	"Jewelry": {"parent": HumanMadeObject, "id":"300209286", "label": "Jewelry", "metatype": "work type"},
	"Manuscript": {"parent": HumanMadeObject, "id":"300265483", "label": "Manuscript", "metatype": "work type"},
	"SiteInstallation": {"parent": HumanMadeObject, "id":"300047896", "label": "Site Installation", "metatype": "work type"},
	"Text": {"parent": HumanMadeObject, "id":"300263751", "label": "Text", "metatype": "work type"},
	"Print":  {"parent": HumanMadeObject, "id":"300041273", "label": "Print", "metatype": "work type"},
	"TimeBasedMedia": {"parent": HumanMadeObject, "id":"300185191", "label": "Time-based Media", "metatype": "work type"},
	"Page": {"parent": HumanMadeObject, "id":"300194222", "label": "Page", "metatype": "work type"},
	"Folio": {"parent": HumanMadeObject, "id":"300189604", "label": "Folio", "metatype": "work type"},

	"Folder": {"parent": HumanMadeObject, 'id':"300197602", "label": "Folder", "metatype": "work type"},
	"Box": {"parent": HumanMadeObject, "id":"300045643", "label": "Box", "metatype": "work type"},
	"Envelope": {"parent": HumanMadeObject, "id":"300197601", "label": "Envelope", "metatype": "work type"},
	"Binder": {"parent": HumanMadeObject,"id":"300252990", "label": "Binder", "metatype": "work type"},
	"Case": {"parent": HumanMadeObject,"id":"300045659", "label": "Case", "metatype": "work type"},
	"FlatfileCabinet": {"parent": HumanMadeObject,"id":"300417284", "label": "Flat File Cabinet", "metatype": "work type"},

	"AuctionCatalog": {"parent": HumanMadeObject,"id":"300026068", "label": "Auction Catalog", "metatype": "work type"},
	"ExhibitionCatalog": {"parent": HumanMadeObject,"id":"300026096", "label": "Exhibition Catalog", "metatype": "work type"},
	"AccessionCatalog": {"parent": HumanMadeObject, "id": "300026617", "label": "Accession Catalog", "metatype": "work type"},
	"SalesCatalog": {"parent": HumanMadeObject,"id":"300026074", "label": "Sales Catalog", "metatype": "work type"},
	"LotteryCatalog": {"parent": HumanMadeObject, "id":"300438603", "label": "Lottery Catalog", "metatype": "work type"},

	"FrontPart": {"parent": HumanMadeObject, "id":"300190703", "label": "Front Part", "metatype": "part type"},
	"BackPart": {"parent": HumanMadeObject, "id":"300190692", "label": "Back Part", "metatype": "part type"},
	"TopPart": {"parent": HumanMadeObject, "id":"300190710", "label": "Top Part", "metatype": "part type"},
	"BottomPart": {"parent": HumanMadeObject, "id":"300190695", "label": "Bottom Part", "metatype": "part type"},
	"SidePart": {"parent": HumanMadeObject, "id":"300190706", "label": "Side Part", "metatype": "part type"},
	"SupportPart": {"parent": HumanMadeObject, "id":"300014844", "label": "Support", "metatype": "part type"},
	"FramePart": {"parent": HumanMadeObject, "id":"300404391", "label": "Frame", "metatype": "part type"},
	"MountPart": {"parent": HumanMadeObject, "id":"300131087", "label": "Mount", "metatype": "part type"},
	"PanelPart": {"parent": HumanMadeObject, "id":"300014657", "label": "Panel", "metatype": "part type"},
	"BasePart": {"parent": HumanMadeObject, "id":"300001656", "label": "Base", "metatype": "part type"},

	"StartingPrice": {"parent": MonetaryAmount, "id": "300417242", "label": "Starting Price"},
	"ReservePrice": {"parent": MonetaryAmount, "id": "300417243", "label": "Reserve Price"},
	"EstimatedPrice": {"parent": MonetaryAmount, "id": "300417244", "label": "Estimated Price"},
	"AskingPrice": {"parent": MonetaryAmount, "id": "300417241", "label": "Asking Price"},
	"HammerPrice": {"parent": MonetaryAmount, "id": "300417245", "label": "Hammer Price"},

	"CommissionPayment": {"parent": Payment, "id": "300393199", "label": "Commission"}

}
 
identity_instances = {

	# Common Materials
	"watercolor": {"parent": Material, "id": "300015045", "label": "watercolors"},	
	"oil": {"parent": Material, "id": "300015050", "label": "oil"},
	"tempera": {"parent": Material, "id": "300015062", "label": "tempera"},
	"canvas": {"parent": Material, "id": "300014078", "label": "canvas"},
	"oak": {"parent": Material, "id": "300012264", "label": "oak"},
	"gold leaf": {"parent": Material, "id": "300264831", "label": "gold leaf"},
	"paper": {"parent": Material, "id": "300014109", "label": "paper"},
	"copper": {"parent": Material, "id": "300011020", "label": "copper"},
	"terracotta": {"parent": Material, "id": "300010669", "label": "terracotta"},
	"glass": {"parent": Material, "id": "300010797", "label": "glass"},
	"chalk": {"parent": Material, "id": "300011727", "label": "chalk"},
	"bronze": {"parent": Material, "id": "300010957", "label": "bronze"},
	"marble": {"parent": Material, "id": "300011443", "label": "marble"},
	"albumen silver print": {"parent": Material, "id": "300127121", "label": "albumen silver print"},
	"gelatin silver print": {"parent": Material, "id": "300128695", "label": "gelatin silver print"},
	"silver": {"parent": Material, "id": "300011029", "label": "silver"},
	"synthetic": {"parent": Type, "id": "xxx", "label": "Synthetic Material"},

	# Measurement Units
	"lignes": {"parent": MeasurementUnit, "id": "300435501", "label": "Paris lines"},
	"fr_inches": {"parent": MeasurementUnit, "id": "300435502", "label": "Paris inches"},
	"fr_feet": {"parent": MeasurementUnit, "id":"300435503", "label": "Paris feet"},
	"inches": {"parent": MeasurementUnit, "id": "300379100", "label": "inches"},
	"feet": {"parent": MeasurementUnit, "id":"300379101", "label": "feet"},
	"cm": {"parent": MeasurementUnit, "id": "300379098", "label": "centimeters"},
	"meters": {"parent": MeasurementUnit, "id": "300379099", "label": "meters"},
	"braccia": {"parent": MeasurementUnit, "id": "300404161", "label": "braccia"},
	"ells": {"parent": MeasurementUnit, "id": "300412070", "label": "ells"},
	"percent": {"parent": MeasurementUnit, "id": "300417377", "label": "percent"},
	"numbers": {"parent": MeasurementUnit, "id": "300055665", "label": "numbers"},
	"bytes": {"parent": MeasurementUnit, "id": "300265869", "label": "bytes"},
	"kilobytes": {"parent": MeasurementUnit, "id": "300265870", "label": "kilobytes"},
	"megabytes": {"parent": MeasurementUnit, "id": "300265873", "label": "megabytes"},
	"gigabytes": {"parent": MeasurementUnit, "id": "300265874", "label": "gigabytes"},
	"terabytes": {"parent": MeasurementUnit, "id": "300266477", "label": "terabytes"},
	"partsUnit": {"parent": MeasurementUnit, "id": "300404159", "label": "parts"},
	"pageCount": {"parent": MeasurementUnit, "id": "300194222", "label": "pages"},
	"pixels": {"parent": MeasurementUnit, "id": "300266190", "label": "pixels"},
	"rgb_colorspace": {"parent": MeasurementUnit, "id": "300266239", "label": "rgb"},
	"seconds": {"parent": MeasurementUnit, "id": "300379239", "label": "seconds"},
	"days": {"parent": MeasurementUnit, "id": "300379242", "label": "days"},
	"months": {"parent": MeasurementUnit, "id": "300379245", "label": "months"},
	"years": {"parent": MeasurementUnit, "id": "300379244", "label": "years"},

	# Languages
	"english": {"parent": Language, "id": "300388277", "label": "English"},
	"french": {"parent": Language, "id":"300388306","label": "French"},
	"german": {"parent": Language, "id":"300388344","label": "German"},
	"spanish": {"parent": Language, "id":"300389311","label": "Spanish"},
	"chinese": {"parent": Language, "id":"300388113", "label":"Chinese"},
	"hindi": {"parent": Language, "id":"300388412", "label":"Hindi"},
	"afrikaans": {"parent": Language, "id":"300387782", "label":"Afrikaans"},
	"albanian": {"parent": Language, "id":"300387803", "label":"Albanian"},
	"arabic": {"parent": Language, "id":"300387843", "label":"Arabic"},
	"bulgarian": {"parent": Language, "id":"300388034", "label":"Bulgarian"},
	"bosnian": {"parent": Language, "id":"300388023", "label":"Bosnian"},
	"catalan": {"parent": Language, "id":"300388072", "label":"Catalan"},
	"portuguese":{"parent": Language, "id":"300389115", "label":"Portuguese"},
	"bengali": {"parent": Language, "id":"300387971", "label":"Bengali"},
	"russian": {"parent": Language, "id":"300389168", "label":"Russian"},
	"dutch": {"parent": Language, "id":"300388256", "label":"Dutch"},
	"finnish": {"parent": Language, "id":"300388299", "label":"Finnish"},
	"icelandic": {"parent": Language, "id":"300388449", "label":"Icelandic"},
	"irish": {"parent": Language, "id":"300388467", "label":"Irish"},
	"italian": {"parent": Language, "id":"300388474", "label":"Italian"},
	"farsi": {"parent": Language, "id":"300388296", "label":"Farsi"},
	"greek": {"parent": Language, "id":"300389734", "label":"Greek"},
	"gujarati": {"parent": Language, "id":"300388371", "label":"Gujarati"},
	"ancient greek": {"parent": Language, "id": "300387827", "label": "Ancient Greek"},
	"korean": {"parent": Language, "id":"300388633", "label":"Korean"},
	"latin": {"parent": Language, "id":"300388693", "label":"Latin"},
	"lithuanian": {"parent": Language, "id":"300388723", "label":"Lithuanian"},
	"macedonian": {"parent": Language, "id":"300388760", "label":"Macedonian"},
	"japanese": {"parent": Language, "id":"300388486", "label":"Japanese"},
	"croatian": {"parent": Language, "id": "300388185", "label": "Croatian"},
	"czech": {"parent": Language, "id": "300388191", "label": "Czech"},
	"danish": {"parent": Language, "id": "300388204", "label": "Danish"},
	"greek": {"parent": Language, "id": "300388361", "label": "Greek"},
	"georgian": {"parent": Language, "id": "300388343", "label": "Georgian"},
	"hebrew": {"parent": Language, "id": "300388401", "label": "Hebrew"},
	"hungarian": {"parent": Language, "id": "300388770", "label": "Magyar (Hungarian)"},
	"norwegian": {"parent": Language, "id": "300388992", "label": "Norwegian"},
	"polish": {"parent": Language, "id": "300389109", "label": "Polish"},
	"romanian": {"parent": Language, "id": "300389157", "label": "Romanian"},
	"serbo-croatian": {"parent": Language, "id": "300389248", "label": "Serbo-Croatian"},
	"slovak": {"parent": Language, "id": "300389290", "label": "Slovak"},
	"slovenian": {"parent": Language, "id": "300389291", "label": "Slovenian"},
	"serbian": {"parent": Language, "id": "300389247", "label": "Serbian"},
	"swedish": {"parent": Language, "id": "300389336", "label": "Swedish"},
	"turkish": {"parent": Language, "id": "300389470", "label": "Turkish"},
	"welsh": {"parent": Language, "id": "300389555", "label": "Welsh"},

	# Currencies
	"us dollars": {"parent": Currency, "id":"300411994", "label": "US Dollars"},
	"gb pounds": {"parent": Currency, "id":"300411998", "label": "British Pounds"},
	"gb shillings": {"parent": Currency, "id":"300412167", "label": "British Shillings"},
	"dutch guilder": {"parent": Currency, "id":"300412019", "label": "Dutch Guilder"},
	"fr assignats": {"parent": Currency, "id":"300412157", "label": "French Assignats"},
	"at shillings": {"parent": Currency, "id":"300412158", "label": "Austrian Shillings"},
	"fr ecus": {"parent": Currency, "id":"300412159", "label": "French Ecus"},
	"de florins": {"parent": Currency, "id":"300412160", "label": "German Florins"},
	"gb guineas": {"parent": Currency, "id":"300412163", "label": "British Guineas"},
	"dk kroner": {"parent": Currency, "id":"300412164", "label": "Danish Kroner"},
	"dk rigsdaler": {"parent": Currency, "id":"300412190", "label": "Danish Rigsdaler"},
	"fr livres": {"parent": Currency, "id":"300412165", "label": "French Livres"},
	"fr louis": {"parent": Currency, "id":"300412166", "label": "French Louis coins"},
	"thalers": {"parent": Currency, "id":"300412168", "label": "Thalers"},
	"de reichsmarks": {"parent": Currency, "id":"300412169", "label": "German Reichsmarks"},
	"de marks": {"parent": Currency, "id":"300412017", "label": "German Marks"},
	"reichsthalers": {"parent": Currency, "id":"300412170", "label": "Reichsthalers"},
	"ch francs": {"parent": Currency, "id":"300412001", "label": "Swiss Francs"},
	"fr francs": {"parent": Currency, "id":"300412016", "label": "French Francs"},
	"it lira": {"parent": Currency, "id":"300412015", "label": "Italian Lira"},
	"ducats": {"parent": Currency, "id":"300193382", "label": "Ducats"},
	"groschen": {"parent": Currency, "id":"300191473", "label": "Groschen"},
	"silbergroshen": {"parent": Currency, "id":"300412171", "label": "Silbergroshen"},
	"stuiver": {"parent": Currency, "id":"300412172", "label": "Stuiver"},
	"at kronen": {"parent": Currency, "id":"300435706", "label": "Austrian Kronen"},
	"austro-hungarian kronen": {"parent": Currency, "id":"300435707", "label": "Austro-Hungarian Kronen"},

	# Color Dimensions -- additional Type on the Color Dimension instance
	# See: https://linked.art/model/object/physical/
	"color red": {"parent": Type, "id": "300126225", "label": "Red"},
	"color green": {"parent": Type, "id": "300128438", "label": "Green"},
	"color blue": {"parent": Type, "id": "300129361", "label": "Blue"},
	"color yellow": {"parent": Type, "id": "300127794", "label": "Yellow"},
	"color orange": {"parent": Type, "id": "300126734", "label": "Orange"},
	"color purple": {"parent": Type, "id": "300130257", "label": "Purple"},
	"color brown": {"parent": Type, "id": "300127490", "label": "Brown"},
	"color black": {"parent": Type, "id": "300130920", "label": "Black"},
	"color white": {"parent": Type, "id": "300129784", "label": "White"},
	"color gray": {"parent": Type, "id": "300130811", "label": "Gray"},

	# Techniques -- use technique property, no need to metatype
	"glassblowing": {"parent": Type, "id":"300053932", "label":"Glassblowing"},
	"sculpting": {"parent": Type, "id":"300264383", "label": "Sculpting"},
	"painting": {"parent": Type, "id":"300054216", "label": "Painting"},
	"spraypainting": {"parent": Type, "id":"300053816", "label": "Spraypainting"},

	# Geographic
	"city": {"parent": Type, "id": "300008389", "label": "City"},
	"province": {"parent": Type, "id":"300000774", "label": "Province"},
	"county": {"parent": Type, "id":"300000771", "label": "County"},
	"nation": {"parent": Type, "id": "300128207", "label": "Nation"},
	"sovereign": {"parent": Type, "id": "300232420", "label": "Sovereign State"},
	"building": {"parent": Type, "id": "300004792", "label": "Building"},
	"city block": {"parent": Type, "id": "300008077", "label": "City Block"},

	# dot one qualifiers
	"style of": {"parent": Type, "id": "300404285", "label": "Style Of"},

	# General
	"primary": {"parent": Type, "id": "300404670", "label": "Primary"},
	"artwork": {"parent": Type, "id": "300133025", "label": "Artwork"},
	"work type": {"parent": Type, "id": "300435443", "label": "Type of Work"},
	"part type": {"parent": Type, "id":"300241583", "label": "Part Type"},
	"brief text": {"parent": Type, "id": "300418049", "label":"Brief Text"},
	"first": {"parent": Type, "id":"300404050", "label": "First"},
	"last": {"parent": Type, "id": "XXX", "label": "Last"},

	# Random?
	"public collection": {"parent": Type, "id": "300411912", "label": "Public Collection"},
	"computer generated": {"parent": Type, "id": "300202389", "label": "Computer Generated"},
	"vandalism": {"parent": Type, "id":"300055299", "label": "Vandalism"},
	"contribution": {"parent": Type, "id":"300403975", "label":"Contribution"},  # As opposed to primarily responsible

	# Subjects -- use is_about / subject project, no need to metatype
	"gender issues": {"parent": Type, "id": "300233686", "label": "Gender Issues"},
	"war": {"parent": Type, "id": "300055314", "label": "War"},
	"crime": {"parent": Type, "id": "300055289", "label": "Crime"},

	# Additional classifications for objects
	"door piece": {"parent": Type, "id":"300002911", "label": "Door piece"}, # on Painting
	"bust": {"parent": Type, "id":"300047457", "label": "Bust"}, # on Sculpture
	"altarpiece": {"parent": Type, "id":"300075940", "label": "Altarpiece"}, # on Painting
	"double-sided paintings": {"parent": Type, "id":"300265620", "label": "Double-sided Paintings"}, # on Painting
	"bas-relief": {"parent": Type, "id":"300184633", "label": "Bas-relief"},

	# Additional classifications for visual content
	"allusion": {"parent": Type, "id":"300055815", "label":"Allusion"},
	"allegory": {"parent": Type, "id":"300055866", "label":"Allegory"}

}

for (name,v) in identity_instances.items():
	register_instance(name, v)
for (name,v) in ext_classes.items():
	register_vocab_class(name, v)


local_instances = {

	# Nationalities
	"american nationality": {"parent": Nationality, "id": "300107956", "label": "American"},
	"austrian nationality": {"parent": Nationality, "id": "300111153", "label": "Austrian"},
	"belgian nationality": {"parent": Nationality, "id": "300111156", "label": "Belgian"},
	"british nationality": {"parent": Nationality, "id": "300111159", "label": "British"},	
	"canadian nationality": {"parent": Nationality, "id": "300107962", "label": "Canadian"},
	"chinese nationality": {"parent": Nationality, "id": "300018322", "label": "Chinese"},
	"czech nationality": {"parent": Nationality, "id": "300111166", "label": "Czech"},
	"danish nationality": {"parent": Nationality, "id": "300111172", "label": "Danish"},
	"dutch nationality": {"parent": Nationality, "id": "300111175", "label": "Dutch"},
	"egyptian nationality": {"parent": Nationality, "id": "300020251", "label": "Egyptian"},
	"flemish nationality": {"parent": Nationality, "id": "300111184", "label": "Flemish"},
	"french nationality": {"parent": Nationality, "id": "300111188", "label": "French"},
	"german nationality": {"parent": Nationality, "id": "300111192", "label": "German"},
	"greek nationality": {"parent": Nationality, "id": "300264816", "label": "Greek"},
	"hungarian nationality": {"parent": Nationality, "id": "300111195", "label": "Hungarian"},
	"indian nationality": {"parent": Nationality, "id": "300018863", "label": "Indian"},
	"irish nationality": {"parent": Nationality, "id": "300111259", "label": "Irish"},
	"italian nationality": {"parent": Nationality, "id": "300111198", "label": "Italian"},
	"japanese nationality": {"parent": Nationality, "id": "300018519", "label": "Japanese"},
	"mexican nationality": {"parent": Nationality, "id": "300107963", "label": "Mexican"},
	"norwegian nationality": {"parent": Nationality, "id": "300111201", "label": "Norwegian"},
	"polish nationality": {"parent": Nationality, "id": "300111204", "label": "Polish"},
	"portuguese nationality": {"parent": Nationality, "id": "300111207", "label": "Portuguese"},
	"russian nationality": {"parent": Nationality, "id": "300111276", "label": "Russian"},
	"spanish nationality": {"parent": Nationality, "id": "300111215", "label": "Spanish"},
	"swedish nationality": {"parent": Nationality, "id": "300111218", "label": "Swedish"},
	"swiss nationality": {"parent": Nationality, "id": "300111221", "label": "Swiss"},


	# Nations
	"belgium": {"parent": Nation, "id": "1000063", "vocab": "tgn", "label": "Belgium"},
	"france": {"parent": Nation, "id": "1000070", "vocab": "tgn", "label": "France"},
	"germany": {"parent": Nation, "id": "7000084", "vocab": "tgn", "label": "Germany"},
	"switzerland": {"parent": Nation, "id": "7011731", "vocab": "tgn", "label": "Switzerland"},
	"united kingdom": {"parent": Nation, "id": "7008591", "vocab": "tgn", "label": "United Kingdom"},
	"england": {"parent": Nation, "id": "7002445", "vocab": "tgn", "label": "England"},
	"scotland": {"parent": Nation, "id": "7002444", "vocab": "tgn", "label": "Scotland"},
	"wales": {"parent": Nation, "id": "7002443", "vocab": "tgn", "label": "Wales"},

	# Shapes
	"oval": {"parent": Shape, "id":"300263817", "label": "Oval"},
	"octagonal": {"parent": Shape, "id":"300263824", "label": "Octagonal"},
	"circle": {"parent": Shape, "id":"300263827", "label": "Circle"},
	"square": {"parent": Shape, "id":"300263832", "label": "Square"},
	"oblong": {"parent": Shape, "id":"300311843", "label": "Oblong"},
	"upright": {"parent": Shape, "id":"300343370", "label": "Upright"},
	"arched": {"parent": Shape, "id":"300126995", "label": "Arched"},

	# Styles
	"style genre": {"parent": Style, "id":"300139140", "label": "Genre"},
	"style landscape": {"parent": Style, "id":"300015636", "label": "Landscape"},
	"style portrait": {"parent": Style, "id":"300015637", "label": "Portrait"},
	"style still life": {"parent": Style, "id":"300015638", "label": "Still Life"},
	"style impressionism": {"parent": Style, "id":"300021503", "label": "Impressionism"},

	# Genders / Biological Sexes
	"male": {"parent": Gender, "id":"300189559", "label": "Male"},
	"female": {"parent": Gender, "id":"300189557", "label": "Female"},	
	"intersex": {"parent": Gender, "id":"300417544", "label": "Intersex"},
	"hermaphrodite": {"parent": Gender, "id":"300389935", "label": "Hermaphrodite"},
	"non-binary gender": {"parent": Gender, "id":"300417543", "label": "Non-Binary Gender Identity"},
	"transgender": {"parent": Gender, "id":"300400504", "label": "Transgender"}

}

for (name,v) in local_instances.items():
	register_instance(name, v)

# pen, pencil, card, cardboard, porcelain, wax, ceramic, plaster
# crayon, millboard, gouache, brass, stone, lead, iron, clay,
# alabaster, limestone

aat_culture_mapping = {
	"italian": "300111198",
	"german": "300111192",
	"dutch": "300020929"
}

def add_classification(obj, cl_type):
	c = cl_type()
	for cn in c._classification:
		if hasattr(obj, 'classified_as') and not cn in obj.classified_as:
			obj.classified_as = cn
		elif not hasattr(obj, 'classified_as'):
			obj.classified_as = cn
	return obj

def make_multitype_obj(*args, **kw):
	# (class1, class2, class3, name=foo, other=bar)
	inst = args[0](**kw)
	for c in args[1:]:
		for cn in c._classification:
			if hasattr(inst, 'classified_as') and not cn in inst.classified_as:
				inst.classified_as = cn
			elif not hasattr(inst, 'classified_as'):
				inst.classified_as = cn
	return inst


def conceptual_only_parts():
	# Make .part work as expected for Right
	# which is only propositional and not symbolic, so P148 not P106

	def set_c_part(self, value):
		self.c_part = value
	def set_c_part_of(self, value):
		self.c_part_of = value

	def rights_getter(self, what):
		if what == "part":
			return self.c_part
		elif what == "part_of":
			return self.c_part_of
		else:
			object.__getattr__(self, what)

	Right.set_part = set_c_part
	Right.set_part_of = set_c_part_of
	Right._property_name_map['c_part'] = 'part'
	Right._property_name_map['c_part_of'] = 'part_of'
	Right._all_properties['part'] = PropositionalObject._all_properties['c_part']
	Right._all_properties['part_of'] = PropositionalObject._all_properties['c_part_of']
	Right.__getattr__ = rights_getter


def add_art_setter():
	# Linked.Art profile requires aat:300133025 on all artworks
	# Art can be a HumanMadeObject or an InformationObject
	# set it by adding art=1 to the constructor

	def art_post_init(self, **kw):
		super(HumanMadeObject, self)._post_init(**kw)
		if "art" in kw:
			self.classified_as = instances['artwork']
	HumanMadeObject._post_init = art_post_init

	def art2_post_init(self, **kw):
		if "art" in kw:
			self.classified_as = instances['artwork']
		super(InformationObject, self)._post_init(**kw)
	InformationObject._post_init = art2_post_init

def add_attribute_assignment_check():
	# Allow references to properties in p177 on AttrAssign
	# Validate that the property is allowed in assigned
	# either on set, or when assigned is set
		
	p177 = factory.context_rev.get('crm:P177_assigned_property', 'assigned_property')
	ass = factory.context_rev.get('crm:P141_assigned', 'assigned')
	assto = factory.context_rev.get('crm:P140:assigned_attribute_to', 'assigned_to')

	phase_rel = factory.context_rev.get('la:relationship', 'relationship')
	phase_of = factory.context_rev.get('la:phase_of', 'phase_of')
	phase_entity = factory.context_rev.get('la:related_entity', 'related_entity')

	def aa_set_assigned(self, value):
		assto_res = getattr(self, assto, None)
		if assto_res:
			p177_res = getattr(self, p177, None)
			assto_res._check_prop(p177_res, value)
		object.__setattr__(self, ass, value)
	setattr(AttributeAssignment, "set_%s" % ass, aa_set_assigned)

	def aa_set_assigned_to(self, value):
		ass_res = getattr(self, ass, None)
		p177_res = getattr(self, p177, None)		
		if ass_res and p177_res:
			# unmap the URI to property name
			value._check_prop(p177_res, ass_res)
		object.__setattr__(self, assto, value)
	setattr(AttributeAssignment, "set_%s" % assto, aa_set_assigned_to)

	def aa_set_assigned_property_type(self, value):
		ass_res = getattr(self, ass, None)
		assto_res = getattr(self, assto, None)
		if ass_res and assto_res:
			assto_res._check_prop(value, ass_res)
		object.__setattr__(self, p177, value)
	setattr(AttributeAssignment, "set_%s" % p177, aa_set_assigned_property_type)


	def phase_set_relationship(self, value):
		# XXX do same checking as above
		object.__setattr__(self, phase_rel, value)
	setattr(Phase, "set_%s" % phase_rel, phase_set_relationship)		

def add_linked_art_boundary_check():

	boundary_classes = [x.__name__ for x in [Actor, HumanMadeObject, Person, Group, VisualItem, \
		Place, Period, LinguisticObject, Phase, Set, Event]]
	data_embed_classes = [Name, Identifier, Dimension, TimeSpan, MonetaryAmount]
	type_embed_classes = [Type, Currency, Language, Material, MeasurementUnit]
	event_embed_classes = [Birth, Creation, Production, Formation, Payment, \
							Death, Destruction, Dissolution ]
	all_embed_classes = []
	all_embed_classes.extend(data_embed_classes)
	all_embed_classes.extend(type_embed_classes)
	all_embed_classes.extend(event_embed_classes)
	embed_classes = [x.__name__ for x in all_embed_classes]

	# Activity, AttributeAssignment, InformationObject, TransferOfCustody, Move
	# Propositional Object

	def my_linked_art_boundary_check(self, top, rel, value):
		# True = Embed ; False = Split
		if isinstance(value, LinguisticObject) and hasattr(value, 'classified_as'):
			for ca in value.classified_as:
				if instances['brief text'] in getattr(ca, 'classified_as', []):
					return True
		# Non Statement Linguistic objects might still be internal or external
		# so apply logic from relating properties, not return False
		elif isinstance(value, ProvenanceEntry):
			return False

		if rel in ["part", "member"]:
			# Downwards, internal simple partitioning 
			# This catches an internal part to a LinguisticObject
			return True
		elif rel in ["part_of", 'member_of']:
			# upwards partition refs are inclusion, and always boundary crossing
			return False
		elif value.type in boundary_classes:
			# This catches the external text LinguisticObject
			return False
		elif value.type in embed_classes:
			return True
		else:
			# Default to embedding to avoid data loss
			return True

	setattr(BaseResource, "_linked_art_boundary_okay", my_linked_art_boundary_check)
	factory.linked_art_boundaries = True	

